from typing import List, Dict, Optional, Union, Any, Iterator, Mapping

from .base import APIBase, Item, ItemList


GET_LIMIT = 20_000
POST_LIMIT = 20_000
PUT_LIMIT = 20_000
PATCH_LIMIT = 20_000


class Production(APIBase):
    ######
    # URLs
    ######

    def get_company_monthly_productions_url(self, filters: Optional[Dict[str, str]] = None) -> str:
        """
        Returns the API url for company monthly production.
        """
        url = f'{self.API_BASE_URL}/monthly-productions'
        if filters is None:
            return url

        url += self._build_params_string(filters)
        return url


    def get_company_daily_productions_url(self, filters: Optional[Dict[str, str]] = None) -> str:
        """
        Returns the API url for company daily production.
        """
        url = f'{self.API_BASE_URL}/daily-productions'
        if filters is None:
            return url

        url += self._build_params_string(filters)
        return url


    def get_project_monthly_productions_url(self, project_id: str, filters: Optional[Dict[str, str]] = None) -> str:
        """
        Returns the API url for a specific project's monthly production.
        """
        url = f'{self.API_BASE_URL}/projects/{project_id}/monthly-productions'
        if filters is None:
            return url

        url += self._build_params_string(filters)
        return url


    def get_project_daily_productions_url(self, project_id: str, filters: Optional[Dict[str, str]] = None) -> str:
        """
        Returns the API url for a specific project's daily production.
        """
        url = f'{self.API_BASE_URL}/projects/{project_id}/daily-productions'
        if filters is None:
            return url

        url += self._build_params_string(filters)
        return url


    ###########
    # API calls
    ###########


    def get_company_monthly_productions(self, filters: Optional[Dict[str, str]] = None) -> ItemList:
        """
        Returns a list of company monthly production items.

        https://docs.api.combocurve.com/#6cecaf35-e501-4e21-899f-22261de76fff
        """
        url = self.get_company_monthly_productions_url(filters)
        params = {'take': GET_LIMIT}
        monthly_production = self._get_items(url, params)

        order = {
            'well': 0,
            'date': 1,
        }
        return self._keysort(monthly_production, order)


    def post_company_monthly_productions(self, data: ItemList) -> ItemList:
        """
        Creates monthly production items.

        https://docs.api.combocurve.com/#8e46ba1e-c230-4d19-bd94-45af8fbe89cb
        """
        url = self.get_company_monthly_productions_url()
        monthly_production = self._post_items(url, data)

        return monthly_production


    def put_company_monthly_productions(self, data: ItemList) -> ItemList:
        """
        Upserts monthly production items.

        https://docs.api.combocurve.com/#7cc3b686-2f9f-470b-8dd0-810fd4fa13dc
        """
        url = self.get_company_monthly_productions_url()
        monthly_production = self._put_items(url, data)

        return monthly_production


    def patch_company_monthly_productions(self, data: ItemList) -> ItemList:
        """
        Updates monthly production items.

        https://docs.api.combocurve.com/#5167e26e-de2d-4860-aa3d-b9b9f80d0d24
        """
        url = self.get_company_monthly_productions_url()
        monthly_production = self._put_items(url, data)

        return monthly_production


    def delete_company_monthly_productions(self, data: ItemList) -> ItemList:
        """
        Deletes monthly production items.

        https://docs.api.combocurve.com/#81ca806f-dfd5-4af2-b155-05439bdf1158
        """
        url = self.get_company_monthly_productions_url()
        monthly_production = self._delete_items(url, data)

        return monthly_production


    def get_company_daily_productions(self, filters: Optional[Dict[str, str]] = None) -> ItemList:
        """
        Returns a list of company monthly production items.

        https://docs.api.combocurve.com/#d53b02b3-6574-40ee-bdfb-1e64191ae80a
        """
        url = self.get_company_daily_productions_url(filters)
        params = {'take': GET_LIMIT}
        dailiy_production = self._get_items(url, params)

        order = {
            'well': 0,
            'date': 1,
        }
        return self._keysort(dailiy_production, order)


    def post_company_daily_productions(self, data: ItemList) -> ItemList:
        """
        Creates daily production items.

        https://docs.api.combocurve.com/#9d7db301-2512-491d-8a8b-985cd3417d3e
        """
        url = self.get_company_daily_productions_url()
        daily_production = self._post_items(url, data)

        return daily_production


    def put_company_daily_productions(self, data: ItemList) -> ItemList:
        """
        Upserts daily production items.

        https://docs.api.combocurve.com/#dfb57b4d-60aa-4116-b219-b174f01cc5dd
        """
        url = self.get_company_daily_productions_url()
        daily_production = self._put_items(url, data)

        return daily_production


    def patch_company_daily_productions(self, data: ItemList) -> ItemList:
        """
        Updates daily production items.

        https://docs.api.combocurve.com/#64460347-9135-4020-9bc6-bb5eeebcbc85
        """
        url = self.get_company_daily_productions_url()
        daily_production = self._patch_items(url, data)

        return daily_production


    def delete_company_daily_productions(self, data: ItemList) -> ItemList:
        """
        Delete daily production items.

        https://docs.api.combocurve.com/#4292e628-bfaa-4b9e-8ef2-467715e1092c
        """
        url = self.get_company_daily_productions_url()
        daily_production = self._delete_items(url, data)

        return daily_production


    def get_project_monthly_productions(self, project_id: str, filters: Optional[Dict[str, str]] = None) -> ItemList:
        """
        Returns a list of monthly production items for a specific project id.

        https://docs.api.combocurve.com/#91a94565-b7cb-40f2-b5a2-0689d8efb897
        """
        url = self.get_project_monthly_productions_url(project_id, filters)
        params = {'take': GET_LIMIT}
        monthly_production = self._get_items(url, params)

        order = {
            'well': 0,
            'date': 1,
        }
        return self._keysort(monthly_production, order)


    def post_project_monthly_productions(self, project_id: str, data: ItemList) -> ItemList:
        """
        Creates project monthly production items.

        https://docs.api.combocurve.com/#2a58c950-1747-43d9-b0d7-c847db39c850
        """
        url = self.get_project_monthly_productions_url(project_id)
        monthly_production = self._post_items(url, data)

        return monthly_production


    def put_project_monthly_productions(self, project_id: str, data: ItemList) -> ItemList:
        """
        Upserts project monthly production items.

        https://docs.api.combocurve.com/#43c7de35-b397-464c-98c6-a5da1b37de5c
        """
        url = self.get_project_monthly_productions_url(project_id)
        monthly_production = self._put_items(url, data)

        return monthly_production


    def patch_project_monthly_productions(self, project_id: str, data: ItemList) -> ItemList:
        """
        Updates project monthly production items.

        https://docs.api.combocurve.com/#b5c7dc84-6897-4650-bf89-2709fd9189a7
        """
        url = self.get_project_monthly_productions_url(project_id)
        monthly_production = self._put_items(url, data)

        return monthly_production


    def delete_project_monthly_productions(self, project_id: str, data: ItemList) -> ItemList:
        """
        Deletes project monthly production items.

        https://docs.api.combocurve.com/#1c826ba7-ff89-434f-bf87-7e3957b25353
        """
        url = self.get_project_monthly_productions_url(project_id)
        monthly_production = self._delete_items(url, data)

        return monthly_production


    def get_project_daily_productions(self, project_id: str, filters: Optional[Dict[str, str]] = None) -> ItemList:
        """
        Returns a list of daily production items for a specific project id.

        https://docs.api.combocurve.com/#637b21b2-6c45-4d19-829c-ffc0514ed86c
        """
        url = self.get_project_daily_productions_url(project_id, filters)
        params = {'take': GET_LIMIT}
        daily_production = self._get_items(url, params)

        order = {
            'well': 0,
            'date': 1,
        }
        return self._keysort(daily_production, order)


    def post_project_daily_productions(self, project_id: str, data: ItemList) -> ItemList:
        """
        Creates project daily production items.

        https://docs.api.combocurve.com/#8e46ba1e-c230-4d19-bd94-45af8fbe89cb
        """
        url = self.get_project_daily_productions_url(project_id)
        daily_production = self._post_items(url, data)

        return daily_production


    def put_project_daily_productions(self, project_id: str, data: ItemList) -> ItemList:
        """
        Upserts project daily production items.

        https://docs.api.combocurve.com/#c0a0e460-0692-4231-a4bc-e2602f73dd3a
        """
        url = self.get_project_daily_productions_url(project_id)
        daily_production = self._put_items(url, data)

        return daily_production


    def patch_project_daily_productions(self, project_id: str, data: ItemList) -> ItemList:
        """
        Updates project daily production items.

        https://docs.api.combocurve.com/#91720250-6958-43e8-9d66-85498a628a02
        """
        url = self.get_project_daily_productions_url(project_id)
        daily_production = self._put_items(url, data)

        return daily_production


    def delete_project_daily_productions(self, project_id: str, data: ItemList) -> ItemList:
        """
        Deletes project daily production items.

        https://docs.api.combocurve.com/#e61b507c-76cb-4e2a-95ba-0c1381938129
        """
        url = self.get_project_daily_productions_url(project_id)
        daily_production = self._delete_items(url, data)

        return daily_production


monthly_production_response = """
        Example response:
        [
            {
                "choke": 13,
                "co2Injection": 123,
                "createdAt": "2020-01-21T16:58:08.986Z",
                "date": "2017-01-15T00:00:00.000Z",
                "customNumber0": 123,
                "customNumber1": 123,
                "customNumber2": 123,
                "customNumber3": 123,
                "customNumber4": 123,
                "daysOn": 30,
                "gas": 16925,
                "gasInjection": 123,
                "ngl": 123,
                "oil": 1022,
                "operationalTag": "Start of production",
                "steamInjection": 123,
                "updatedAt": "2020-01-21T17:58:08.986Z",
                "water": 4631,
                "waterInjection": 123,
                "well": "5e286e0d8f3df51d7ae744a3"
            },
            {
                "choke": 13,
                "co2Injection": 123,
                "createdAt": "2020-01-21T16:58:08.986Z",
                "date": "2017-01-15T00:00:00.000Z",
                "customNumber0": 123,
                "customNumber1": 123,
                "customNumber2": 123,
                "customNumber3": 123,
                "customNumber4": 123,
                "daysOn": 30,
                "gas": 16925,
                "gasInjection": 123,
                "ngl": 123,
                "oil": 1022,
                "operationalTag": "Start of production",
                "steamInjection": 123,
                "updatedAt": "2020-01-21T17:58:08.986Z",
                "water": 4631,
                "waterInjection": 123,
                "well": "5e286e0d8f3df51d7ae744a3"
            }
        ]
"""

monthly_post_put_patch_response = """
        Example data:
        [
            {
                "date": "<date>",
                "bottomHolePressure": "<number>",
                "casingHeadPressure": "<number>",
                "choke": "<number>",
                "chosenID": "<string>",
                "co2Injection": "<number>",
                "customNumber0": "<number>",
                "customNumber1": "<number>",
                "customNumber2": "<number>",
                "customNumber3": "<number>",
                "customNumber4": "<number>",
                "dataSource": "<string>",
                "flowlinePressure": "<number>",
                "gas": "<number>",
                "gasInjection": "<number>",
                "gasLiftInjectionPressure": "<number>",
                "hoursOn": "<number>",
                "ngl": "<number>",
                "oil": "<number>",
                "operationalTag": "<string>",
                "steamInjection": "<number>",
                "tubingHeadPressure": "<number>",
                "vesselSeparatorPressure": "<number>",
                "water": "<number>",
                "waterInjection": "<number>",
                "well": "<string>"
            },
            {
                "date": "<date>",
                "bottomHolePressure": "<number>",
                "casingHeadPressure": "<number>",
                "choke": "<number>",
                "chosenID": "<string>",
                "co2Injection": "<number>",
                "customNumber0": "<number>",
                "customNumber1": "<number>",
                "customNumber2": "<number>",
                "customNumber3": "<number>",
                "customNumber4": "<number>",
                "dataSource": "<string>",
                "flowlinePressure": "<number>",
                "gas": "<number>",
                "gasInjection": "<number>",
                "gasLiftInjectionPressure": "<number>",
                "hoursOn": "<number>",
                "ngl": "<number>",
                "oil": "<number>",
                "operationalTag": "<string>",
                "steamInjection": "<number>",
                "tubingHeadPressure": "<number>",
                "vesselSeparatorPressure": "<number>",
                "water": "<number>",
                "waterInjection": "<number>",
                "well": "<string>"
            }
        ]

        Example response:
        [
            {
                "generalErrors": [
                    {
                        "name": "ValidationError",
                        "message": "The field 'well' is required.",
                        "location": "In body of request at position [0]"
                    },
                    {
                        "name": "ValidationError",
                        "message": "The field 'date' is required.",
                        "location": "In body of request at position [2]"
                    }
                ],
                "results": [
                    {
                        "status": "Success",
                        "code": 200,
                        "well": "62b1c13e2750169012ee4515",
                        "date": "2023-01-01T00:00:00.000Z"
                    },
                    {
                        "status": "Success",
                        "code": 200,
                        "well": "62b1c13e4857169000ee4613",
                        "date": "2023-01-01T00:00:00.000Z"
                    }
                ],
                "failedCount": 2,
                "successCount": 2
            }
        ]
"""

daily_production_response = """
        Example data:
        [
            {
                "date": "<date>",
                "bottomHolePressure": "<number>",
                "casingHeadPressure": "<number>",
                "choke": "<number>",
                "chosenID": "<string>",
                "co2Injection": "<number>",
                "customNumber0": "<number>",
                "customNumber1": "<number>",
                "customNumber2": "<number>",
                "customNumber3": "<number>",
                "customNumber4": "<number>",
                "dataSource": "<string>",
                "flowlinePressure": "<number>",
                "gas": "<number>",
                "gasInjection": "<number>",
                "gasLiftInjectionPressure": "<number>",
                "hoursOn": "<number>",
                "ngl": "<number>",
                "oil": "<number>",
                "operationalTag": "<string>",
                "steamInjection": "<number>",
                "tubingHeadPressure": "<number>",
                "vesselSeparatorPressure": "<number>",
                "water": "<number>",
                "waterInjection": "<number>",
                "well": "<string>"
            },
            {
                "date": "<date>",
                "bottomHolePressure": "<number>",
                "casingHeadPressure": "<number>",
                "choke": "<number>",
                "chosenID": "<string>",
                "co2Injection": "<number>",
                "customNumber0": "<number>",
                "customNumber1": "<number>",
                "customNumber2": "<number>",
                "customNumber3": "<number>",
                "customNumber4": "<number>",
                "dataSource": "<string>",
                "flowlinePressure": "<number>",
                "gas": "<number>",
                "gasInjection": "<number>",
                "gasLiftInjectionPressure": "<number>",
                "hoursOn": "<number>",
                "ngl": "<number>",
                "oil": "<number>",
                "operationalTag": "<string>",
                "steamInjection": "<number>",
                "tubingHeadPressure": "<number>",
                "vesselSeparatorPressure": "<number>",
                "water": "<number>",
                "waterInjection": "<number>",
                "well": "<string>"
            }
        ]

        Example response:
        [
            {
                "generalErrors": [
                    {
                        "name": "ValidationError",
                        "message": "The field 'well' is required.",
                        "location": "In body of request at position [0]"
                    },
                    {
                        "name": "ValidationError",
                        "message": "The field 'date' is required.",
                        "location": "In body of request at position [2]"
                    }
                ],
                "results": [
                    {
                        "status": "Success",
                        "code": 200,
                        "well": "62b1c13e2750169012ee4515",
                        "date": "2023-01-01T00:00:00.000Z"
                    },
                    {
                        "status": "Success",
                        "code": 200,
                        "well": "62b1c13e4857169000ee4613",
                        "date": "2023-01-01T00:00:00.000Z"
                    }
                ],
                "failedCount": 2,
                "successCount": 2
            }
        ]
"""

daily_post_put_patch_response = """
        Example data:
        [
            {
                "date": "<date>",
                "bottomHolePressure": "<number>",
                "casingHeadPressure": "<number>",
                "choke": "<number>",
                "chosenID": "<string>",
                "co2Injection": "<number>",
                "customNumber0": "<number>",
                "customNumber1": "<number>",
                "customNumber2": "<number>",
                "customNumber3": "<number>",
                "customNumber4": "<number>",
                "dataSource": "<string>",
                "flowlinePressure": "<number>",
                "gas": "<number>",
                "gasInjection": "<number>",
                "gasLiftInjectionPressure": "<number>",
                "hoursOn": "<number>",
                "ngl": "<number>",
                "oil": "<number>",
                "operationalTag": "<string>",
                "steamInjection": "<number>",
                "tubingHeadPressure": "<number>",
                "vesselSeparatorPressure": "<number>",
                "water": "<number>",
                "waterInjection": "<number>",
                "well": "<string>"
            },
            {
                "date": "<date>",
                "bottomHolePressure": "<number>",
                "casingHeadPressure": "<number>",
                "choke": "<number>",
                "chosenID": "<string>",
                "co2Injection": "<number>",
                "customNumber0": "<number>",
                "customNumber1": "<number>",
                "customNumber2": "<number>",
                "customNumber3": "<number>",
                "customNumber4": "<number>",
                "dataSource": "<string>",
                "flowlinePressure": "<number>",
                "gas": "<number>",
                "gasInjection": "<number>",
                "gasLiftInjectionPressure": "<number>",
                "hoursOn": "<number>",
                "ngl": "<number>",
                "oil": "<number>",
                "operationalTag": "<string>",
                "steamInjection": "<number>",
                "tubingHeadPressure": "<number>",
                "vesselSeparatorPressure": "<number>",
                "water": "<number>",
                "waterInjection": "<number>",
                "well": "<string>"
            }
        ]

        Example response:
        [
            {
                "generalErrors": [
                    {
                        "name": "ValidationError",
                        "message": "The field 'well' is required.",
                        "location": "In body of request at position [0]"
                    },
                    {
                        "name": "ValidationError",
                        "message": "The field 'date' is required.",
                        "location": "In body of request at position [2]"
                    }
                ],
                "results": [
                    {
                        "status": "Success",
                        "code": 200,
                        "well": "62b1c13e2750169012ee4515",
                        "date": "2023-01-01T00:00:00.000Z"
                    },
                    {
                        "status": "Success",
                        "code": 200,
                        "well": "62b1c13e4857169000ee4613",
                        "date": "2023-01-01T00:00:00.000Z"
                    }
                ],
                "failedCount": 2,
                "successCount": 2
            }
        ]
"""


Production.get_company_monthly_productions.__doc__ += monthly_production_response  # type: ignore [operator]
Production.get_project_monthly_productions.__doc__ += monthly_production_response  # type: ignore [operator]

Production.get_company_daily_productions.__doc__ += daily_production_response  # type: ignore [operator]
Production.get_project_daily_productions.__doc__ += daily_production_response  # type: ignore [operator]

Production.post_company_monthly_productions.__doc__ += monthly_post_put_patch_response  # type: ignore [operator]
Production.put_company_monthly_productions.__doc__ += monthly_post_put_patch_response  # type: ignore [operator]
Production.patch_company_monthly_productions.__doc__ += monthly_post_put_patch_response  # type: ignore [operator]

Production.post_company_daily_productions.__doc__ += daily_post_put_patch_response  # type: ignore [operator]
Production.put_company_daily_productions.__doc__ += daily_post_put_patch_response  # type: ignore [operator]
Production.patch_company_daily_productions.__doc__ += daily_post_put_patch_response  # type: ignore [operator]



from typing import List, Dict, Optional, Union, Any, Iterator, Mapping

from .base import APIBase, Item, ItemList


GET_LIMIT = 200


class Projects(APIBase):
    ######
    # URLs
    ######

    def get_projects_url(self, filters: Optional[Dict[str, str]] = None) -> str:
        """
        Returns the API url for projects.
        """
        url = f'{self.API_BASE_URL}/projects'
        if filters is None:
            return url

        url += self._build_params_string(filters)
        return url


    def get_project_by_id_url(self, project_id: str) -> str:
        """
        Returns the API url for a specific project from its project id.
        """
        base_url = self.get_projects_url()
        return f'{base_url}/{project_id}'


    ###########
    # API calls
    ###########


    def get_projects(self, filters: Optional[Dict[str, str]] = None) -> ItemList:
        """
        Returns a list of projects.

        https://docs.api.combocurve.com/#25b8ab6b-5c03-49bc-ba57-06f063b31a97

        Example response:
        [
            {
                "createdAt": "2020-01-21T16:58:08.986Z",
                "id": "5e5981b9e23dae0012624d72",
                "name": "Test project",
                "updatedAt": "2020-01-21T17:58:08.986Z"
            }
        ]
        """
        url = self.get_projects_url(filters)
        params = {'take': GET_LIMIT}
        projects = self._get_items(url, params=params)

        order = {
            'name': 0,
            'id': 3,
            'createdAt': 2,
            'updatedAt': 1,
        }
        return self._keysort(projects, order)


    def post_projects(self, data: ItemList) -> ItemList:
        """
        Creates a new project.

        https://docs.api.combocurve.com/#a89937e7-92ed-4184-a4c0-9009732cd4e8

        Example request:
        [
            {
                "name": "Test project"
            }
        ]

        Example response:
        [
            {
                "generalErrors": [
                    {
                        "name": "ValidationError",
                        "message": "The field 'id' is required.",
                        "location": "In body of request at position [0]"
                    },
                    {
                        "name": "ValidationError",
                        "message": "The field 'dataSource' is required.",
                        "location": "In body of request at position [2]"
                    }
                ],
                "results": [
                    {
                        "status": "Success",
                        "code": 200,
                        "name": "Acme Royalties 2021-10-28",
                        "id": "61698aa08eca904d9cc5b622"
                    },
                    {
                        "status": "Success",
                        "code": 200,
                        "name": "Acme Royalties 2021-10-28",
                        "id": "61698aa08eca904d9cc5b623"
                    }
                ],
                "failedCount": 2,
                "successCount": 2
            }
        ]
        """
        url = self.get_projects_url()
        projects = self._post_items(url, data)

        return projects


    def get_project_by_id(self, id: str) -> Item:
        """
        Returns a specific project from its project id.

        https://docs.api.combocurve.com/#3d58f1b6-41f5-4bf0-9d79-613f14d08bed

        Example response:
        {
            "createdAt": "2020-01-21T16:58:08.986Z",
            "id": "5e5981b9e23dae0012624d72",
            "name": "Test project",
            "updatedAt": "2020-01-21T17:58:08.986Z"
        }
        """
        url = self.get_project_by_id_url(id)
        projects = self._get_items(url)

        return projects[0]



__version__ = '1.1.0'

from .root import Root
from .projects import Projects
from .scenarios import Scenarios
from .production import Production
from .econ_runs import EconRuns
from .wells import Wells
from .models import Models
from .forecasts import Forecasts
from .typecurves import TypeCurves
from .directional import Directional

from .base import Item, ItemList, PrimativeValue, IterableValue


class ComboCurveAPI(
    Root,
    Projects,
    Scenarios,
    Production,
    EconRuns,
    Wells,
    Models,
    Forecasts,
    TypeCurves,
    Directional
):
    """
    This class is the primary interface for interacting with the Combo Curve
    API. It inherits all of the API endpoints from the other classes in this
    module. It is intended to be used as a single entrypoint for interacting
    with the ComboCurve API.
    """
    pass



import requests  # type: ignore
import warnings

from combocurve_api_v1.pagination import get_next_page_url  # type: ignore

from typing import List, Dict, Optional, Union, Any, Iterator, Mapping, cast

from .base import APIBase, Item, ItemList


GET_LIMIT = 200


class Scenarios(APIBase):
    ######
    # URLs
    ######

    def get_scenarios_url(self, project_id: str, filters: Optional[Dict[str, str]] = None) -> str:
        """
        Returns the API url for project scenarios scoped from the project's id.
        """
        url = f'{self.API_BASE_URL}/projects/{project_id}/scenarios'
        if filters is None:
            return url

        url += self._build_params_string(filters)
        return url


    def get_scenario_by_id_url(self, project_id: str, scenario_id: str) -> str:
        """
        Returns the API url for a specific project scenario from its
        scenario id.
        """
        base_url = self.get_scenarios_url(project_id)
        return f'{base_url}/{scenario_id}'


    def get_scenario_combos_url(
            self, project_id: str, scenario_id: str, filters: Optional[Dict[str, str]] = None) -> str:
        """
        Returns the API url for combos for a specific project id, scenario id,
        and econ run id.
        """
        base_url = self.get_scenario_by_id_url(project_id, scenario_id)
        url = f'{base_url}/combos'
        if filters is None:
            return url

        url += self._build_params_string(filters)
        return url



    def get_scenario_qualifiers_url(
            self, project_id: str, scenario_id: str, econ_name: Optional[str] = None,
            filters: Optional[Dict[str, str]] = None) -> str:
        """
        Returns the API url for qualifiers for a specific project id,
        scenario id, and optionally, econ name.
        """
        base_url = self.get_scenario_by_id_url(project_id, scenario_id)
        url = f'{base_url}/qualifiers'

        if filters is None:
            filters = {}

        if econ_name is not None:
            for model in self.ECON_MODELS:
                if econ_name.casefold() == model['econModelType'].casefold():
                    filters['econName'] = econ_name
                    break

            else:
                model_names = [m['econModelType'] for m in self.ECON_MODELS]
                warnings.warn(
                    f'`econ_name` is not in list of valid names:\n{model_names}. All qualifiers will be returned.')

        url += self._build_params_string(filters)
        return url


    def get_scenario_wells_url(
            self, project_id: str, scenario_id: str, filters: Optional[Dict[str, str]] = None) -> str:
        """
        Returns the API url for well assignments for a specific project id and
        scenario id.
        """
        base_url = self.get_scenario_by_id_url(project_id, scenario_id)
        url = f'{self.API_BASE_URL}/projects/{project_id}/scenarios/{scenario_id}/well-assignments'
        if filters is None:
            return url

        url += self._build_params_string(filters)
        return url


    ###########
    # API calls
    ###########


    # Scenarios


    def get_scenarios(self, project_id: str, filters: Optional[Dict[str, str]] = None) -> ItemList:
        """
        Returns a list of scenarios scoped from the project's id.

        https://docs.api.combocurve.com/#e7de7ef5-228b-4de6-bd67-89c8ef14a4bc
        """
        url = self.get_scenarios_url(project_id, filters)
        params = {'take': GET_LIMIT}
        scenarios = self._get_items(url, params=params)

        order = {
            'name': 0,
            'id': 3,
            'createdAt': 2,
            'updatedAt': 1,
        }
        return self._keysort(scenarios, order)


    def post_scenarios(self, project_id: str, data: ItemList) -> ItemList:
        """
        Creates scenarios for a specific project id.

        https://docs.api.combocurve.com/#52df57aa-af96-4e86-976c-8d4aff3125f7
        """
        url = self.get_scenarios_url(project_id)
        scenarios = self._post_items(url, data)

        return scenarios


    def put_scenarios(self, project_id: str, data: ItemList) -> ItemList:
        """
        Upserts scenarios for a specific project id.

        https://docs.api.combocurve.com/#94b372d8-6696-4f65-b1b7-6c0576b3f75b
        """
        url = self.get_scenarios_url(project_id)
        scenarios = self._put_items(url, data)

        return scenarios


    def delete_scenarios(self, project_id: str, scenario_name: Optional[str], scenario_id: Optional[str]) -> ItemList:
        """
        Deletes scenarios for a specific project id.

        https://docs.api.combocurve.com/#7429717e-51f2-48de-ad22-6c4bc5bf1bb1

        Returns the headers from the delete response where 'X-Delete-Count' is
        the number of wells deleted.
        """
        if (scenario_name or scenario_id) is None:
            raise ValueError('Must provide at least one of scenario_name or scenario_id')

        filters: Dict[str, str] = {}
        if scenario_name is not None:
            filters['name'] = scenario_name

        if scenario_id is not None:
            filters['id'] = scenario_id

        url = self.get_scenarios_url(project_id, filters)
        scenarios = self._delete_responses(url, data=[])

        headers = scenarios[0].headers

        return headers


    def get_scenario_by_id(self, project_id: str, scenario_id: str) -> Item:
        """
        Returns a specific scenario from its scenario id.

        https://docs.api.combocurve.com/#e7de7ef5-228b-4de6-bd67-89c8ef14a4bc

        Example response:
        [
            {
                "createdAt": "2020-01-21T16:58:08.986Z",
                "id": "5e5981b9e23dae0012624d72",
                "name": "Test scenario",
                "updatedAt": "2020-01-21T17:58:08.986Z"
            }
        ]
        """
        url = self.get_scenario_by_id_url(project_id, scenario_id)
        scenarios = self._get_items(url)

        return scenarios[0]


    # Combos


    def get_scenario_combos(self, project_id: str, scenario_id: str) -> ItemList:
        """
        Returns a list of combos for a specific project id and scenario id.

        https://docs.api.combocurve.com/#844e5b66-d20f-48ab-a186-380a1cc49630
        """
        url = self.get_scenario_combos_url(project_id, scenario_id)
        params = {'take': GET_LIMIT}
        return self._get_items(url, params)


    def post_scenario_combos(self, project_id: str, scenario_id: str, data: ItemList) -> ItemList:
        """
        Creates scenario combos for a specific project id and scenario id.

        https://docs.api.combocurve.com/#e14969da-bd78-4747-9f19-e0277836dfee
        """
        url = self.get_scenario_combos_url(project_id, scenario_id)
        scenarios = self._post_items(url, data)

        return scenarios


    def put_scenario_combos(self, project_id: str, scenario_id: str, data: ItemList) -> ItemList:
        """
        Upserts scenario combos for a specific project id and scenario id.

        https://docs.api.combocurve.com/#2e47dad1-c176-4628-b97f-8ace00c6ad0d
        """
        url = self.get_scenario_combos_url(project_id, scenario_id)
        scenarios = self._put_items(url, data)

        return scenarios


    def delete_scenario_combo(self, project_id: str, scenario_id: str, saved_name: str) -> ItemList:
        """
        Deletes scenario combos for a specific project id and scenario id.

        https://docs.api.combocurve.com/#881f45fd-287f-4f6c-a805-2e690d675a7a

        Returns the headers from the delete response where 'X-Delete-Count' is
        the number of wells deleted.
        """
        filters: Dict[str, str] = {}
        filters['savedName'] = saved_name

        url = self.get_scenario_combos_url(project_id, scenario_id, filters)
        scenarios = self._delete_responses(url, data=[])

        headers = scenarios[0].headers

        return headers


    # Qualifiers


    def get_scenario_qualifiers(self, project_id: str, scenario_id: str, econ_name: Optional[str] = None) -> Item:
        """
        Returns a list of qualifiers for a specific project id, scenario id and
        econ name.

        https://docs.api.combocurve.com/#76132581-aefd-4efa-ae82-2af8596340de
        """
        url = self.get_scenario_qualifiers_url(project_id, scenario_id, econ_name)
        qualifiers = self._get_items(url)

        return qualifiers[0]


    def post_scenario_qualifiers(self, project_id: str, scenario_id: str, data: ItemList) -> ItemList:
        """
        Creates scenario qualifiers for a specific project id and scenario id.

        https://docs.api.combocurve.com/#b917f0ea-9bf2-4fea-8c16-d780ade2ac2d
        """
        url = self.get_scenario_qualifiers_url(project_id, scenario_id)
        scenarios = self._post_items(url, data)

        return scenarios


    def put_scenario_qualifiers(self, project_id: str, scenario_id: str, data: ItemList) -> ItemList:
        """
        Upserts scenario qualifiers for a specific project id and scenario id.

        https://docs.api.combocurve.com/#f2d33571-1d1b-4f7b-b8a7-02d437dd4f02
        """
        url = self.get_scenario_qualifiers_url(project_id, scenario_id)
        scenarios = self._put_items(url, data)

        return scenarios


    def delete_scenario_qualifiers(
            self, project_id: str, scenario_id: str,
            econ_names: str, qualifier_names: str) -> ItemList:
        """
        Deletes scenario qualifiers for a specific project id and scenario id.

        https://docs.api.combocurve.com/#75353547-a52c-444d-953f-219a0a006a51

        Returns the headers from the delete response where 'X-Delete-Count' is
        the number of wells deleted.
        """
        filters: Dict[str, str] = {'qualifierNames': qualifier_names}

        url = self.get_scenario_qualifiers_url(project_id, scenario_id, econ_names, filters)
        scenarios = self._delete_responses(url, data=[])

        headers = scenarios[0].headers

        return headers

    # Scenario Wells


    def get_scenario_wells(self, project_id: str, scenario_id: str) -> ItemList:
        """
        Returns a list of well assignments for a specific project id and
        scenario id.

        https://docs.api.combocurve.com/#7849574f-3530-4751-a21a-485e694609e6
        """
        url = self.get_scenario_wells_url(project_id, scenario_id)
        return self._get_items(url)


    def post_scenario_wells(self, project_id: str, scenario_id: str, data: ItemList) -> ItemList:
        """
        Creates scenario well assignments for a specific project id and scenario id.

        https://docs.api.combocurve.com/#538885a6-3b78-44aa-9852-5b5574bd60f3
        """
        url = self.get_scenario_wells_url(project_id, scenario_id)
        scenarios = self._post_items(url, data)

        return scenarios


    def put_scenario_wells(self, project_id: str, scenario_id: str, data: ItemList) -> ItemList:
        """
        Upserts scenario well assignments for a specific project id and scenario id.

        https://docs.api.combocurve.com/#b4f4a676-b3aa-4207-ae06-69b282da36fd
        """
        url = self.get_scenario_wells_url(project_id, scenario_id)
        scenarios = self._put_items(url, data)

        return scenarios


    def delete_scenario_wells(self, project_id: str, scenario_id: str, wells: str) -> ItemList:
        """
        Deletes scenario well assignments for a specific project id and scenario id.

        https://docs.api.combocurve.com/#96bc7d64-831c-44ff-a707-4f0b31bb6c34

        Returns the headers from the delete response where 'X-Delete-Count' is
        the number of wells deleted.
        """
        filters: Dict[str, str] = {}
        filters['wells'] = wells

        url = self.get_scenario_wells_url(project_id, scenario_id, filters)
        scenarios = self._delete_responses(url, data=[])

        headers = scenarios[0].headers

        return headers


post_put_scenarios_response = """
        Example data:
        [
            {
                "id": "<string>",
                "name": "<string>"
            },
            {
                "id": "<string>",
                "name": "<string>"
            }
        ]

        Example response:
        [
            {
                "generalErrors": [
                    {
                        "name": "ValidationError",
                        "message": "The field 'name' is required.",
                        "location": "[0]"
                    },
                    {
                        "name": "ValidationError",
                        "message": "The field 'unique' is required.",
                        "location": "[2]"
                    }
                ],
                "results": [
                    {
                        "status": "Success",
                        "code": 200,
                        "chosenID": "5e5981b9e23dae0012624d72"
                    }
                ],
                "failedCount": 2,
                "successCount": 2
            }
        ]
"""

post_put_scenario_combos_response = """
        Example data:
        [
            {
                "savedName": "<string>",
                "combos": [
                    {
                        "comboName": "<string>",
                        "qualifiers": [
                            {
                                "assumption": "<string>",
                                "qualifierName": "<string>"
                            },
                            {
                                "assumption": "<string>",
                                "qualifierName": "<string>"
                            }
                        ],
                        "selected": "<boolean>"
                    },
                    {
                        "comboName": "<string>",
                        "qualifiers": [
                            {
                                "assumption": "<string>",
                                "qualifierName": "<string>"
                            },
                            {
                                "assumption": "<string>",
                                "qualifierName": "<string>"
                            }
                        ],
                        "selected": "<boolean>"
                    }
                ]
            },
            {
                "savedName": "<string>",
                "combos": [
                    {
                        "comboName": "<string>",
                        "qualifiers": [
                            {
                                "assumption": "<string>",
                                "qualifierName": "<string>"
                            },
                            {
                                "assumption": "<string>",
                                "qualifierName": "<string>"
                            }
                        ],
                        "selected": "<boolean>"
                    },
                    {
                        "comboName": "<string>",
                        "qualifiers": [
                            {
                                "assumption": "<string>",
                                "qualifierName": "<string>"
                            },
                            {
                                "assumption": "<string>",
                                "qualifierName": "<string>"
                            }
                        ],
                        "selected": "<boolean>"
                    }
                ]
            }
        ]

        Example response:
        [
            {
                "failedCount": 1,
                "generalErrors": [
                    {
                        "chosenID": "chosen_id",
                        "location": ".person.name",
                        "message": "The field name is required",
                        "name": "ValidationError"
                    }
                ],
                "results": [
                    {
                        "chosenID": "chosen_id",
                        "code": 201,
                        "status": "created"
                    }
                ],
                "successCount": 1
            }
        ]
"""

post_put_scenario_qualifiers_response = """
        Example data:
        [
            {
                "econModel": "<string>",
                "name": "<string>",
                "newName": "<string>"
            },
            {
                "econModel": "<string>",
                "name": "<string>",
                "newName": "<string>"
            }
        ]

        Example response:
        [
            {
                "failedCount": 1,
                "generalErrors": [
                    {
                        "chosenID": "chosen_id",
                        "location": ".person.name",
                        "message": "The field name is required",
                        "name": "ValidationError"
                    }
                ],
                "results": [
                    {
                        "chosenID": "chosen_id",
                        "code": 201,
                        "status": "created"
                    }
                ],
                "successCount": 1
            }
        ]
"""

post_put_scenario_wells_response = """
        Example data:
        [
            "<string>",
            "<string>"
        ]

        Example response:
        [
            {
                "failedCount": 1,
                "generalErrors": [
                    {
                        "chosenID": "chosen_id",
                        "location": ".person.name",
                        "message": "The field name is required",
                        "name": "ValidationError"
                    }
                ],
                "results": [
                    {
                        "chosenID": "chosen_id",
                        "code": 201,
                        "status": "created"
                    }
                ],
                "successCount": 1
            }
        ]
"""

Scenarios.post_scenarios.__doc__ += post_put_scenarios_response  # type: ignore [operator]
Scenarios.put_scenarios.__doc__ += post_put_scenarios_response  # type: ignore [operator]

Scenarios.post_scenario_combos.__doc__ += post_put_scenario_combos_response  # type: ignore [operator]
Scenarios.put_scenario_combos.__doc__ += post_put_scenario_combos_response  # type: ignore [operator]

Scenarios.post_scenario_qualifiers.__doc__ += post_put_scenario_qualifiers_response  # type: ignore [operator]
Scenarios.put_scenario_qualifiers.__doc__ += post_put_scenario_qualifiers_response  # type: ignore [operator]

Scenarios.post_scenario_wells.__doc__ += post_put_scenario_wells_response  # type: ignore [operator]
Scenarios.put_scenario_wells.__doc__ += post_put_scenario_wells_response  # type: ignore [operator]



from typing import List, Dict, Optional, Union, Any, Iterator, Mapping, TypedDict, cast

from .base import APIBase, Item, ItemList


GET_LIMIT = 200


class Root(APIBase):
    ######
    # URLs
    ######

    def get_custom_columns_url(self, collection: str, filters: Optional[Dict[str, str]] = None) -> str:
        """
        Returns the API url for custom columns.
        """
        url = f'{self.API_BASE_URL}/custom-columns/{collection}'
        if filters is None:
            return url

        url += self._build_params_string(filters)
        return url


    def get_well_identifiers_url(self) -> str:
        """
        Returns the API url for well identifiers.
        """
        return f'{self.API_BASE_URL}/well-identifiers'


    def get_tags_url(self, filters: Optional[Dict[str, str]] = None) -> str:
        """
        Returns the API url for tags.
        """
        url = f'{self.API_BASE_URL}/tags'
        if filters is None:
            return url

        url += self._build_params_string(filters)
        return url


    def get_root_econ_runs_url(self, filters: Optional[Dict[str, str]] = None) -> str:
        """
        Returns the API url for econ runs.
        """
        url = f'{self.API_BASE_URL}/econ-runs'
        if filters is None:
            return url

        url += self._build_params_string(filters)
        return url


    def get_root_econ_run_by_id_url(self, id: str) -> str:
        """
        Returns the API url for a specific econ run from its econ run id.
        """
        return f'{self.API_BASE_URL}/econ-runs/{id}'


    def get_root_forecast_daily_volumes_url(self, filters: Optional[Dict[str, str]] = None) -> str:
        """
        Returns the API url for daily volumes.
        """
        url = f'{self.API_BASE_URL}/forecasts/daily-volumes'
        if filters is None:
            return url

        url += self._build_params_string(filters)
        return url


    def get_root_forecast_monthly_volumes_url(self, filters: Optional[Dict[str, str]] = None) -> str:
        """
        Returns the API url for monthly volumes.
        """
        url = f'{self.API_BASE_URL}/forecasts/monthly-volumes'
        if filters is None:
            return url

        url += self._build_params_string(filters)
        return url


    ###########
    # API calls
    ###########


    def get_custom_columns(self, collection: str, filters: Optional[Dict[str, str]] = None) -> ItemList:
        """
        Returns a list of custom columns.

        https://docs.api.combocurve.com/#3c047d2a-db2c-419f-9187-1a4db81215eb
        """
        url = self.get_custom_columns_url(collection, filters)
        return self._get_items(url)


    def patch_well_identifiers(self, data: ItemList) -> ItemList:
        """
        Update well identifiers.

        https://docs.api.combocurve.com/#7d7f9a19-e693-4b59-9769-71efd58cba31

        Structure of data:
        [
            ...,
            {
                "wellId": "5e272d39b78210dd2a1bd8fe", // required
                        "newInfo": { // at least one of them.
                            "chosenKeyID": "api14",
                            "companyScope": true,
                            "dataSource": "internal"
                        }
            },
            ...
        ]
        """
        url = self.get_well_identifiers_url()
        return self._patch_items(url, data)


    def get_tags(self, filters: Optional[Dict[str, str]] = None) -> ItemList:
        """
        Returns a list of tags.

        https://docs.api.combocurve.com/#17a1bfdc-e23e-4b4a-a033-4dc31fb4e35a

        Example response:
        [
            {
                "createdAt": "2021-07-27T17:52:28.791Z",
                "name": "Test tag",
                "description": "Test tag description",
                "updatedAt": "2021-07-27T17:52:28.791Z"
            },
            {
                "createdAt": "2021-07-27T17:52:28.791Z",
                "name": "Test tag 2",
                "description": "Test tag 2 description",
                "updatedAt": "2021-07-27T17:52:28.791Z"
            }
        ]
        """
        url = self.get_tags_url(filters)
        params = {'take': GET_LIMIT}
        return self._get_items(url, params)


    def get_root_econ_runs(self, filters: Optional[Dict[str, str]] = None) -> ItemList:
        """
        Returns a list of econ runs.

        https://docs.api.combocurve.com/#489ddb62-4cdd-4470-a572-00dc0e10e73b
        """
        url = self.get_root_econ_runs_url(filters)
        params = {'take': GET_LIMIT}
        return self._get_items(url, params)


    def get_root_econ_run_by_id(self, id: str) -> Item:
        """
        Returns a specific econ run from its econ run id.

        https://docs.api.combocurve.com/#bba4ff40-fd07-4ce2-a2c7-368a97e99e93
        """
        url = self.get_root_econ_run_by_id_url(id)
        params = {'take': GET_LIMIT}
        return self._get_items(url, params)[0]


    def get_root_forecast_daily_volumes(self, filters: Optional[Dict[str, str]] = None) -> ItemList:
        """
        Returns a list of daily volumes.

        https://docs.api.combocurve.com/#351629d8-78f0-459c-9463-ba4ffb9675d3
        """
        url = self.get_root_forecast_daily_volumes_url(filters)
        params = {'take': GET_LIMIT}
        return self._get_items(url, params)


    def get_root_forecast_monthly_volumes(self, filters: Optional[Dict[str, str]] = None) -> ItemList:
        """
        Returns a list of monthly volumes.

        https://docs.api.combocurve.com/#26cc17ce-3c64-44f0-82c8-e3fe9243e191
        """
        url = self.get_root_forecast_monthly_volumes_url(filters)
        params = {'take': GET_LIMIT}
        return self._get_items(url, params)



from typing import List, Dict, Optional, Union, Any, Iterator, Mapping

from .base import APIBase, Item, ItemList


GET_LIMIT = 200
POST_LIMIT = 500
PUT_LIMIT = 500

SORT_ORDER = {
    'name': 0,
    'id': 3,
    'createdAt': 2,
    'updatedAt': 1,
}


class Models(APIBase):
    ######
    # URLs
    ######


    def get_econ_models_url(self, project_id: str, filters: Optional[Dict[str, str]] = None) -> str:
        """
        Returns the API url of econ models for a specific project id.
        """
        url = f'{self.API_BASE_URL}/projects/{project_id}/econ-models'
        if filters is None:
            return url

        url += self._build_params_string(filters)
        return url


    def get_econ_models_by_type_url(
            self, project_id: str, econ_model_type: str, filters: Optional[Dict[str, str]] = None) -> str:
        """
        Returns the API url of econ models for a specific project id and model
        type. Allows `econModelType` passed as a parameter rather than calling
        a different function for each model type.
        """
        def _get_route_for_model(econ_model_type: str) -> Union[str, None]:
            for model in APIBase.ECON_MODELS:
                if model['econModelType'].casefold() == econ_model_type.casefold():
                    return model['route']

            return None

        route = _get_route_for_model(econ_model_type)
        assert route is not None, f'Invalid econ model type: {econ_model_type}'
        url = f'{self.API_BASE_URL}/projects/{project_id}/econ-models/{route}'
        if filters is None:
            return url

        url += self._build_params_string(filters)
        return url


    def get_econ_model_by_type_by_id_url(
            self, project_id: str, econ_model_type: str, model_id: str,
            filters: Optional[Dict[str, str]] = None) -> str:
        """
        Returns the API url of a sepcific econ model for a specific project id
        and model type. Allows `econModelType` passed as a parameter rather
        than calling a different function for each model type.
        """
        base_url = self.get_econ_models_by_type_url(project_id, econ_model_type)
        url = f'{base_url}/{model_id}'
        if filters is None:
            return url

        url += self._build_params_string(filters)
        return url


    def get_econ_model_assignments_by_type_by_id_url(
            self, project_id: str, econ_model_type: str, model_id: str,
            filters: Optional[Dict[str, str]] = None) -> str:
        """
        Returns the API url for assignments of a sepcific econ model for a
        specific project id and model type. Allows `econModelType` passed as a
        parameter rather than calling a different function for each model type.
        """
        def _get_route_for_assignment(econ_model_type: str) -> Union[str, None]:
            for model in APIBase.ECON_MODELS:
                if model['econModelType'].casefold() == econ_model_type.casefold():
                    return model['econModelType']

            return None

        route = _get_route_for_assignment(econ_model_type)
        assert route is not None, f'Invalid econ model type: {econ_model_type}'
        url = f'{self.API_BASE_URL}/projects/{project_id}/econ-models/{route}/{model_id}/assignments'
        if filters is None:
            return url

        url += self._build_params_string(filters)
        return url


    def get_general_options_models_url(self, project_id: str, filters: Optional[Dict[str, str]] = None) -> str:
        """
        Returns the API url of general options models for a specific project id.
        """
        econ_model_type = 'GeneralOptions'
        return self.get_econ_models_by_type_url(project_id, econ_model_type, filters)


    def get_general_options_model_by_id_url(self, project_id: str, model_id: str) -> str:
        """
        Returns the API url of a specific general options model for a specific
        project id.
        """
        econ_model_type = 'GeneralOptions'
        return self.get_econ_model_by_type_by_id_url(project_id, econ_model_type, model_id)


    def get_actual_forecast_models_url(self, project_id: str, filters: Optional[Dict[str, str]] = None) -> str:
        """
        Returns the API url of actual-forecast models for a specific project id.
        """
        econ_model_type = 'ActualOrForecast'
        return self.get_econ_models_by_type_url(project_id, econ_model_type, filters)


    def get_actual_forecast_model_by_id_url(self, project_id: str, model_id: str) -> str:
        """
        Returns the API url of a specific actual-forecast model for a specific
        project id.
        """
        econ_model_type = 'ActualOrForecast'
        return self.get_econ_model_by_type_by_id_url(project_id, econ_model_type, model_id)


    def get_actual_forecast_assignments_by_id_url(self, project_id: str, model_id: str) -> str:
        """
        Returns the API url of a specific actual-forecast model assignment for
        a specific project id.
        """
        econ_model_type = 'ActualOrForecast'
        return self.get_econ_model_assignments_by_type_by_id_url(project_id, econ_model_type, model_id)


    def get_capex_models_url(self, project_id: str, filters: Optional[Dict[str, str]] = None) -> str:
        """
        Returns the API url of capex models for a specific project id.
        """
        econ_model_type = 'Capex'
        return self.get_econ_models_by_type_url(project_id, econ_model_type, filters)


    def get_capex_model_by_id_url(self, project_id: str, model_id: str) -> str:
        """
        Returns the API url of a specific capex model for a specific
        project id.
        """
        econ_model_type = 'Capex'
        return self.get_econ_model_by_type_by_id_url(project_id, econ_model_type, model_id)


    def get_capex_assignments_by_id_url(self, project_id: str, model_id: str) -> str:
        """
        Returns the API url of a specific capex model assignment for a specific
        project id.
        """
        econ_model_type = 'Capex'
        return self.get_econ_model_assignments_by_type_by_id_url(project_id, econ_model_type, model_id)


    def get_date_settings_models_url(self, project_id: str, filters: Optional[Dict[str, str]] = None) -> str:
        """
        Returns the API url of date settings models for a specific project id.
        """
        econ_model_type = 'Dates'
        return self.get_econ_models_by_type_url(project_id, econ_model_type, filters)

    def get_date_settings_model_by_id_url(self, project_id: str, model_id: str) -> str:
        """
        Returns the API url of a specific date settings model for a specific
        project id.
        """
        econ_model_type = 'Dates'
        return self.get_econ_model_by_type_by_id_url(project_id, econ_model_type, model_id)


    def get_date_settings_assignments_by_id_url(self, project_id: str, model_id: str) -> str:
        """
        Returns the API url of a specific date settings model assignment for a
        specific project id.
        """
        econ_model_type = 'Dates'
        return self.get_econ_model_assignments_by_type_by_id_url(project_id, econ_model_type, model_id)


    def get_depreciation_models_url(self, project_id: str, filters: Optional[Dict[str, str]] = None) -> str:
        """
        Returns the API url of depreciation models for a specific project id.
        """
        econ_model_type = 'Depreciation'
        return self.get_econ_models_by_type_url(project_id, econ_model_type, filters)


    def get_depreciation_model_by_id_url(self, project_id: str, model_id: str) -> str:
        """
        Returns the API url of a specific depreciation model for a specific
        project id.
        """
        econ_model_type = 'Depreciation'
        return self.get_econ_model_by_type_by_id_url(project_id, econ_model_type, model_id)


    def get_depreciation_assignments_by_id_url(self, project_id: str, model_id: str) -> str:
        """
        Returns the API url of a specific depreciation model assignment for a
        specific project id.
        """
        econ_model_type = 'Depreciation'
        return self.get_econ_model_assignments_by_type_by_id_url(project_id, econ_model_type, model_id)


    def get_differentials_models_url(self, project_id: str, filters: Optional[Dict[str, str]] = None) -> str:
        """
        Returns the API url of differentials models for a specific project id.
        """
        econ_model_type = 'Differentials'
        return self.get_econ_models_by_type_url(project_id, econ_model_type, filters)


    def get_differentials_model_by_id_url(self, project_id: str, model_id: str) -> str:
        """
        Returns the API url of a specific differentials model for a specific
        project id.
        """
        econ_model_type = 'Differentials'
        return self.get_econ_model_by_type_by_id_url(project_id, econ_model_type, model_id)


    def get_differentials_assignments_by_id_url(self, project_id: str, model_id: str) -> str:
        """
        Returns the API url of a specific differentials model assignment for a
        specific project id.
        """
        econ_model_type = 'Differentials'
        return self.get_econ_model_assignments_by_type_by_id_url(project_id, econ_model_type, model_id)


    def get_emissions_models_url(self, project_id: str, filters: Optional[Dict[str, str]] = None) -> str:
        """
        Returns the API url of emissions models for a specific project id.
        """
        econ_model_type = 'Emission'
        return self.get_econ_models_by_type_url(project_id, econ_model_type, filters)


    def get_emissions_model_by_id_url(self, project_id: str, model_id: str) -> str:
        """
        Returns the API url of a specific emissions model for a specific
        project id.
        """
        econ_model_type = 'Emission'
        return self.get_econ_model_by_type_by_id_url(project_id, econ_model_type, model_id)


    def get_emissions_assignments_by_id_url(self, project_id: str, model_id: str) -> str:
        """
        Returns the API url of a specific emissions model assignment for a
        specific project id.
        """
        econ_model_type = 'Emission'
        return self.get_econ_model_assignments_by_type_by_id_url(project_id, econ_model_type, model_id)


    def get_escalations_models_url(self, project_id: str, filters: Optional[Dict[str, str]] = None) -> str:
        """
        Returns the API url of escalations models for a specific project id.
        """
        econ_model_type = 'Escalation'
        return self.get_econ_models_by_type_url(project_id, econ_model_type, filters)


    def get_escalations_model_by_id_url(self, project_id: str, model_id: str) -> str:
        """
        Returns the API url of a specific escalations model for a specific
        project id.
        """
        econ_model_type = 'Escalation'
        return self.get_econ_model_by_type_by_id_url(project_id, econ_model_type, model_id)

    def get_escalations_assignments_by_id_url(self, project_id: str, model_id: str) -> str:
        """
        Returns the API url of a specific escalations model assignment for a
        specific project id.
        """
        econ_model_type = 'Escalation'
        return self.get_econ_model_assignments_by_type_by_id_url(project_id, econ_model_type, model_id)


    def get_expenses_models_url(self, project_id: str, filters: Optional[Dict[str, str]] = None) -> str:
        """
        Returns the API url of expenses models for a specific project id.
        """
        econ_model_type = 'Expenses'
        return self.get_econ_models_by_type_url(project_id, econ_model_type, filters)


    def get_expenses_model_by_id_url(self, project_id: str, model_id: str) -> str:
        """
        Returns the API url of a specific expenses model for a specific
        project id.
        """
        econ_model_type = 'Expenses'
        return self.get_econ_model_by_type_by_id_url(project_id, econ_model_type, model_id)


    def get_expenses_assignments_by_id_url(self, project_id: str, model_id: str) -> str:
        """
        Returns the API url of a specific expenses model assignment for a
        specific project id.
        """
        econ_model_type = 'Expenses'
        return self.get_econ_model_assignments_by_type_by_id_url(project_id, econ_model_type, model_id)


    def get_fluid_models_url(self, project_id: str, filters: Optional[Dict[str, str]] = None) -> str:
        """
        Returns the API url of fluid models for a specific project id.
        """
        econ_model_type = 'FluidModel'
        return self.get_econ_models_by_type_url(project_id, econ_model_type, filters)


    def get_fluid_models_by_id_url(self, project_id: str, model_id: str) -> str:
        """
        Returns the API url of a specific fluid model for a specific
        project id.
        """
        econ_model_type = 'FluidModel'
        return self.get_econ_model_by_type_by_id_url(project_id, econ_model_type, model_id)


    def get_fluid_assignments_by_id_url(self, project_id: str, model_id: str) -> str:
        """
        Returns the API url of a specific fluid model assignment for a specific
        project id.
        """
        econ_model_type = 'FluidModel'
        return self.get_econ_model_assignments_by_type_by_id_url(project_id, econ_model_type, model_id)


    def get_ownership_reversions_models_url(self, project_id: str, filters: Optional[Dict[str, str]] = None) -> str:
        """
        Returns the API url of ownership reversions models for a specific
        project id.
        """
        econ_model_type = 'OwnershipReversion'
        return self.get_econ_models_by_type_url(project_id, econ_model_type, filters)


    def get_ownership_reversions_model_by_id_url(self, project_id: str, model_id: str) -> str:
        """
        Returns the API url of a specific ownership reversions model for a
        specific project id.
        """
        econ_model_type = 'OwnershipReversion'
        return self.get_econ_model_by_type_by_id_url(project_id, econ_model_type, model_id)


    def get_ownership_reversions_assignments_by_id_url(self, project_id: str, model_id: str) -> str:
        """
        Returns the API url of a specific ownership reversions model assignment
        for a specific project id.
        """
        econ_model_type = 'OwnershipReversion'
        return self.get_econ_model_assignments_by_type_by_id_url(project_id, econ_model_type, model_id)


    def get_pricing_models_url(self, project_id: str, filters: Optional[Dict[str, str]] = None) -> str:
        """
        Returns the API url of pricing models for a specific project id.
        """
        econ_model_type = 'Pricing'
        return self.get_econ_models_by_type_url(project_id, econ_model_type, filters)


    def get_pricing_model_by_id_url(self, project_id: str, model_id: str) -> str:
        """
        Returns the API url of a specific pricing model for a specific
        project id.
        """
        econ_model_type = 'Pricing'
        return self.get_econ_model_by_type_by_id_url(project_id, econ_model_type, model_id)


    def get_pricing_assignments_by_id_url(self, project_id: str, model_id: str) -> str:
        """
        Returns the API url of a specific pricing model assignment for a
        specific project id.
        """
        econ_model_type = 'Pricing'
        return self.get_econ_model_assignments_by_type_by_id_url(project_id, econ_model_type, model_id)


    def get_production_taxes_models_url(self, project_id: str, filters: Optional[Dict[str, str]] = None) -> str:
        """
        Returns the API url of production taxes models for a specific
        project id.
        """
        econ_model_type = 'ProductionTaxes'
        return self.get_econ_models_by_type_url(project_id, econ_model_type, filters)


    def get_production_taxes_model_by_id_url(self, project_id: str, model_id: str) -> str:
        """
        Returns the API url of a specific production taxes model for a specific
        project id.
        """
        econ_model_type = 'ProductionTaxes'
        return self.get_econ_model_by_type_by_id_url(project_id, econ_model_type, model_id)

    def get_production_taxes_assignments_by_id_url(self, project_id: str, model_id: str) -> str:
        """
        Returns the API url of a specific production taxes model assignment for
        a specific project id.
        """
        econ_model_type = 'ProductionTaxes'
        return self.get_econ_model_assignments_by_type_by_id_url(project_id, econ_model_type, model_id)


    def get_reserves_categories_models_url(self, project_id: str, filters: Optional[Dict[str, str]] = None) -> str:
        """
        Returns the API url of reserves categories models for a specific
        project id.
        """
        econ_model_type = 'ReservesCategory'
        return self.get_econ_models_by_type_url(project_id, econ_model_type, filters)


    def get_reserves_categories_model_by_id_url(self, project_id: str, model_id: str) -> str:
        """
        Returns the API url of a specific reserves categories model for a
        specific project id.
        """
        econ_model_type = 'ReservesCategory'
        return self.get_econ_model_by_type_by_id_url(project_id, econ_model_type, model_id)


    def get_reserves_categories_assignments_by_id_url(self, project_id: str, model_id: str) -> str:
        """
        Returns the API url of a specific reserves categories model assignment
        for a specific project id.
        """
        econ_model_type = 'ReservesCategory'
        return self.get_econ_model_assignments_by_type_by_id_url(project_id, econ_model_type, model_id)


    def get_riskings_models_url(self, project_id: str, filters: Optional[Dict[str, str]] = None) -> str:
        """
        Returns the API url of riskings models for a specific project id.
        """
        econ_model_type = 'Risking'
        return self.get_econ_models_by_type_url(project_id, econ_model_type, filters)


    def get_riskings_model_by_id_url(self, project_id: str, model_id: str) -> str:
        """
        Returns the API url of a specific riskings model for a specific
        project id.
        """
        econ_model_type = 'Risking'
        return self.get_econ_model_by_type_by_id_url(project_id, econ_model_type, model_id)


    def get_riskings_assignments_by_id_url(self, project_id: str, model_id: str) -> str:
        """
        Returns the API url of a specific riskings model assignment for a
        specific project id.
        """
        econ_model_type = 'Risking'
        return self.get_econ_model_assignments_by_type_by_id_url(project_id, econ_model_type, model_id)


    def get_stream_properties_models_url(self, project_id: str, filters: Optional[Dict[str, str]] = None) -> str:
        """
        Returns the API url of stream properties models for a specific
        project id.
        """
        econ_model_type = 'StreamProperties'
        return self.get_econ_models_by_type_url(project_id, econ_model_type, filters)


    def get_stream_properties_model_by_id_url(self, project_id: str, model_id: str) -> str:
        """
        Returns the API url of a specific stream properties model for a
        specific project id.
        """
        econ_model_type = 'StreamProperties'
        return self.get_econ_model_by_type_by_id_url(project_id, econ_model_type, model_id)


    def get_stream_properties_assignments_by_id_url(self, project_id: str, model_id: str) -> str:
        """
        Returns the API url of a specific stream properties model assignment
        for a specific project id.
        """
        econ_model_type = 'StreamProperties'
        return self.get_econ_model_assignments_by_type_by_id_url(project_id, econ_model_type, model_id)



    ###########
    # API calls
    ###########


    def get_econ_models(self, project_id: str, filters: Optional[Dict[str, str]] = None) -> ItemList:
        """
        Returns a list of econ models.
        """
        url = self.get_econ_models_url(project_id, filters)
        params = {'take': GET_LIMIT}
        econ_models = self._get_items(url, params)

        return self._keysort(econ_models, SORT_ORDER)


    def get_econ_models_by_type(
            self, project_id: str, econ_model_type: str, filters: Optional[Dict[str, str]] = None) -> ItemList:
        """
        Returns a list of econ models by type. Allows `econModelType` passed as
        a parameter rather than calling a different function for each model
        type.
        """
        url = self.get_econ_models_by_type_url(project_id, econ_model_type, filters)
        params = {'take': GET_LIMIT}
        econ_models = self._get_items(url, params)

        return self._keysort(econ_models, SORT_ORDER)


    def get_econ_model_by_type_by_id(
            self, project_id: str, econ_model_type: str, model_id: str) -> Union[Item, None]:
        """
        Returns a specific econ model from its type and id. Allows
        `econModelType` passed as a parameter rather than calling a different
        function for each model type.
        """
        url = self.get_econ_model_by_type_by_id_url(project_id, econ_model_type, model_id)
        econ_model = self._get_items(url)

        return econ_model[0]


    def get_econ_model_assignments_by_type_by_id(
            self, project_id: str, econ_model_type: str, model_id: str) -> Union[ItemList, None]:
        """
        Returns a specific econ model assignment from its type and id. Allows
        `econModelType` passed as a parameter rather than calling a different
        function for each model type.
        """
        url = self.get_econ_model_assignments_by_type_by_id_url(project_id, econ_model_type, model_id)
        assignments = self._get_items(url)

        if len(assignments) == 0:
            return None

        return assignments


    def get_general_options_models(self, project_id: str, filters: Optional[Dict[str, str]] = None) -> ItemList:
        """
        Returns a list of general options models.
        """
        url = self.get_general_options_models_url(project_id, filters)
        params = {'take': GET_LIMIT}
        general_options = self._get_items(url, params)

        return self._keysort(general_options, SORT_ORDER)


    def get_general_options_model_by_id(self, project_id: str, model_id: str) -> Item:
        """
        Returns a specific general options model from its id.
        """
        url = self.get_general_options_model_by_id_url(project_id, model_id)
        general_options = self._get_items(url)

        return general_options[0]


    def get_actual_forecast_models(self, project_id: str, filters: Optional[Dict[str, str]] = None) -> ItemList:
        """
        Returns a list of actual-forecast models.
        """
        url = self.get_actual_forecast_models_url(project_id, filters)
        params = {'take': GET_LIMIT}
        actual_forecast = self._get_items(url, params)

        return self._keysort(actual_forecast, SORT_ORDER)


    def get_actual_forecast_model_by_id(self, project_id: str, model_id: str) -> Item:
        """
        Returns a specific actual-forecast model from its id.
        """
        url = self.get_actual_forecast_model_by_id_url(project_id, model_id)
        actual_forecast = self._get_items(url)

        return actual_forecast[0]


    def get_actual_forecast_assignments_by_id(self, project_id: str, model_id: str) -> Union[ItemList, None]:
        """
        Returns a specific actual-forecast model assignment from its id.
        """
        url = self.get_actual_forecast_assignments_by_id_url(project_id, model_id)
        assignments = self._get_items(url)

        if len(assignments) == 0:
            return None

        return assignments


    def get_reserves_categories_models(self, project_id: str, filters: Optional[Dict[str, str]] = None) -> ItemList:
        """
        Returns a list of reserves categories models.
        """
        url = self.get_reserves_categories_models_url(project_id, filters)
        params = {'take': GET_LIMIT}
        reserves_categories = self._get_items(url, params)

        return self._keysort(reserves_categories, SORT_ORDER)


    def get_reserves_categories_by_id(self, project_id: str, model_id: str) -> Item:
        """
        Returns a specific reserves categories model from its id.
        """
        url = self.get_reserves_categories_model_by_id_url(project_id, model_id)
        reserves_categories = self._get_items(url)

        return reserves_categories[0]


    def get_reserves_categories_assignments_by_id(self, project_id: str, model_id: str) -> Union[ItemList, None]:
        """
        Returns a specific reserves categories model assignment from its id.
        """
        url = self.get_reserves_categories_assignments_by_id_url(project_id, model_id)
        assignments = self._get_items(url)

        if len(assignments) == 0:
            return None

        return assignments


    def get_escalation_models(self, project_id: str, filters: Optional[Dict[str, str]] = None) -> ItemList:
        """
        Returns a list of escalations models.
        """
        url = self.get_escalations_models_url(project_id, filters)
        params = {'take': GET_LIMIT}
        escalations = self._get_items(url, params)

        return self._keysort(escalations, SORT_ORDER)


    def get_escalations_model_by_id(self, project_id: str, model_id: str) -> Item:
        """
        Returns a specific escalations model from its id.
        """
        url = self.get_escalations_model_by_id_url(project_id, model_id)
        escalations = self._get_items(url)

        return escalations[0]


    def get_escalations_assignments_by_id(self, project_id: str, model_id: str) -> Union[ItemList, None]:
        """
        Returns a specific escalations model assignment from its id.
        """
        url = self.get_actual_forecast_assignments_by_id_url(project_id, model_id)
        assignments = self._get_items(url)

        if len(assignments) == 0:
            return None

        return assignments

    def get_differential_models(self, project_id: str, filters: Optional[Dict[str, str]] = None) -> ItemList:
        """
        Returns a list of differentials models.
        """
        url = self.get_differentials_models_url(project_id, filters)
        params = {'take': GET_LIMIT}
        differentials = self._get_items(url, params)

        return self._keysort(differentials, SORT_ORDER)


    def get_differentials_model_by_id(self, project_id: str, model_id: str) -> Item:
        """
        Returns a specific differentials model from its id.
        """
        url = self.get_differentials_model_by_id_url(project_id, model_id)
        differentials = self._get_items(url)

        return differentials[0]


    def get_differentials_assignments_by_id(self, project_id: str, model_id: str) -> Union[ItemList, None]:
        """
        Returns a specific differentials model assignment from its id.
        """
        url = self.get_differentials_assignments_by_id_url(project_id, model_id)
        assignments = self._get_items(url)

        if len(assignments) == 0:
            return None

        return assignments


    def get_pricing_models(self, project_id: str, filters: Optional[Dict[str, str]] = None) -> ItemList:
        """
        Returns a list of pricing models.
        """
        url = self.get_pricing_models_url(project_id, filters)
        params = {'take': GET_LIMIT}
        pricing = self._get_items(url, params)

        return self._keysort(pricing, SORT_ORDER)


    def get_pricing_model_by_id(self, project_id: str, model_id: str) -> Item:
        """
        Returns a specific pricing model from its id.
        """
        url = self.get_pricing_model_by_id_url(project_id, model_id)
        pricing = self._get_items(url)

        return pricing[0]


    def get_pricing_assignments_by_id(self, project_id: str, model_id: str) -> Union[ItemList, None]:
        """
        Returns a specific pricing model assignment from its id.
        """
        url = self.get_pricing_assignments_by_id_url(project_id, model_id)
        assignments = self._get_items(url)

        if len(assignments) == 0:
            return None

        return assignments


    def get_ownership_reversions(self, project_id: str, filters: Optional[Dict[str, str]] = None) -> ItemList:
        """
        Returns a list of ownership reversions models.
        """
        url = self.get_ownership_reversions_models_url(project_id, filters)
        params = {'take': GET_LIMIT}
        ownership_reversions = self._get_items(url, params)

        return self._keysort(ownership_reversions, SORT_ORDER)


    def get_ownership_reversions_model_by_id(self, project_id: str, model_id: str) -> Item:
        """
        Returns a specific ownership reversions model from its id.
        """
        url = self.get_ownership_reversions_model_by_id_url(project_id, model_id)
        ownership_reversions = self._get_items(url)

        return ownership_reversions[0]


    def get_ownership_reversions_assignments_by_id(self, project_id: str, model_id: str) -> Union[ItemList, None]:
        """
        Returns a specific ownership reversions model assignment from its id.
        """
        url = self.get_ownership_reversions_assignments_by_id_url(project_id, model_id)
        assignments = self._get_items(url)

        if len(assignments) == 0:
            return None

        return assignments


    def get_production_taxes_models(self, project_id: str, filters: Optional[Dict[str, str]] = None) -> ItemList:
        """
        Returns a list of production taxes models.
        """
        url = self.get_production_taxes_models_url(project_id, filters)
        params = {'take': GET_LIMIT}
        production_taxes = self._get_items(url, params)

        return self._keysort(production_taxes, SORT_ORDER)


    def get_production_taxes_model_by_id(self, project_id: str, model_id: str) -> Item:
        """
        Returns a specific production taxes model from its id.
        """
        url = self.get_production_taxes_model_by_id_url(project_id, model_id)
        production_taxes = self._get_items(url)

        return production_taxes[0]


    def get_production_taxes_assignments_by_id(self, project_id: str, model_id: str) -> Union[ItemList, None]:
        """
        Returns a specific production taxes model assignment from its id.
        """
        url = self.get_production_taxes_assignments_by_id_url(project_id, model_id)
        assignments = self._get_items(url)

        if len(assignments) == 0:
            return None

        return assignments


    def get_risking_models(self, project_id: str, filters: Optional[Dict[str, str]] = None) -> ItemList:
        """
        Returns a list of riskings models.
        """
        url = self.get_riskings_models_url(project_id, filters)
        params = {'take': GET_LIMIT}
        riskings = self._get_items(url, params)

        return self._keysort(riskings, SORT_ORDER)


    def get_riskings_model_by_id(self, project_id: str, model_id: str) -> Item:
        """
        Returns a specific riskings model from its id.
        """
        url = self.get_riskings_model_by_id_url(project_id, model_id)
        riskings = self._get_items(url)

        return riskings[0]


    def get_riskings_assignments_by_id(self, project_id: str, model_id: str) -> Union[ItemList, None]:
        """
        Returns a specific riskings model assignment from its id.
        """
        url = self.get_riskings_assignments_by_id_url(project_id, model_id)
        assignments = self._get_items(url)

        if len(assignments) == 0:
            return None

        return assignments

    def get_stream_propertie_models(self, project_id: str, filters: Optional[Dict[str, str]] = None) -> ItemList:
        """
        Returns a list of stream properties models.
        """
        url = self.get_stream_properties_models_url(project_id, filters)
        params = {'take': GET_LIMIT}
        stream_properties = self._get_items(url, params)

        return self._keysort(stream_properties, SORT_ORDER)


    def get_stream_properties_model_by_id(self, project_id: str, model_id: str) -> Item:
        """
        Returns a specific stream properties model from its id.
        """
        url = self.get_stream_properties_model_by_id_url(project_id, model_id)
        stream_properties = self._get_items(url)

        return stream_properties[0]


    def get_stream_properties_assignments_by_id(self, project_id: str, model_id: str) -> Union[ItemList, None]:
        """
        Returns a specific stream properties model assignment from its id.
        """
        url = self.get_stream_properties_assignments_by_id_url(project_id, model_id)
        assignments = self._get_items(url)

        if len(assignments) == 0:
            return None

        return assignments

    def get_expense_models(self, project_id: str, filters: Optional[Dict[str, str]] = None) -> ItemList:
        """
        Returns a list of expenses models.
        """
        url = self.get_expenses_models_url(project_id, filters)
        params = {'take': GET_LIMIT}
        expenses = self._get_items(url, params)

        return self._keysort(expenses, SORT_ORDER)


    def get_expenses_model_by_id(self, project_id: str, model_id: str) -> Item:
        """
        Returns a specific expenses model from its id.
        """
        url = self.get_expenses_model_by_id_url(project_id, model_id)
        expenses = self._get_items(url)

        return expenses[0]


    def get_expenses_assignments_by_id(self, project_id: str, model_id: str) -> Union[ItemList, None]:
        """
        Returns a specific expenses model assignment from its id.
        """
        url = self.get_expenses_assignments_by_id_url(project_id, model_id)
        assignments = self._get_items(url)

        if len(assignments) == 0:
            return None

        return assignments


    def get_emission_models(self, project_id: str, filters: Optional[Dict[str, str]] = None) -> ItemList:
        """
        Returns a list of emissions models.
        """
        url = self.get_emissions_models_url(project_id, filters)
        params = {'take': GET_LIMIT}
        emissions = self._get_items(url, params)


        return self._keysort(emissions, SORT_ORDER)


    def get_emissions_model_by_id(self, project_id: str, model_id: str) -> Item:
        """
        Returns a specific emissions model from its id.
        """
        url = self.get_emissions_model_by_id_url(project_id, model_id)
        emissions = self._get_items(url)

        return emissions[0]


    def get_emissions_assignments_by_id(self, project_id: str, model_id: str) -> Union[ItemList, None]:
        """
        Returns a specific emissions model assignment from its id.
        """
        url = self.get_emissions_assignments_by_id_url(project_id, model_id)
        assignments = self._get_items(url)

        if len(assignments) == 0:
            return None

        return assignments


    def get_fluid_models(self, project_id: str, filters: Optional[Dict[str, str]] = None) -> ItemList:
        """
        Returns a list of fluid models.
        """
        url = self.get_fluid_models_url(project_id, filters)
        params = {'take': GET_LIMIT}
        fluid_models = self._get_items(url, params)

        return self._keysort(fluid_models, SORT_ORDER)


    def get_fluid_model_by_id(self, project_id: str, model_id: str) -> Item:
        """
        Returns a specific fluid model from its id.
        """
        url = self.get_fluid_models_by_id_url(project_id, model_id)
        fluid_models = self._get_items(url)

        return fluid_models[0]


    def get_fluid_assignments_by_id(self, project_id: str, model_id: str) -> Union[ItemList, None]:
        """
        Returns a specific fluid model assignment from its id.
        """
        url = self.get_fluid_assignments_by_id_url(project_id, model_id)
        assignments = self._get_items(url)

        if len(assignments) == 0:
            return None

        return assignments


    def get_capex(self, project_id: str, filters: Optional[Dict[str, str]] = None) -> ItemList:
        """
        Returns a list of capex models.
        """
        url = self.get_capex_models_url(project_id, filters)
        params = {'take': GET_LIMIT}
        capex = self._get_items(url, params)

        return self._keysort(capex, SORT_ORDER)


    def get_capex_model_by_id(self, project_id: str, model_id: str) -> Item:
        """
        Returns a specific capex model from its id.
        """
        url = self.get_capex_model_by_id_url(project_id, model_id)
        capex = self._get_items(url)

        return capex[0]


    def get_capex_assignments_by_id(self, project_id: str, model_id: str) -> Union[ItemList, None]:
        """
        Returns a specific capex model assignment from its id.
        """
        url = self.get_capex_assignments_by_id_url(project_id, model_id)
        assignments = self._get_items(url)

        if len(assignments) == 0:
            return None

        return assignments


    def get_date_settings(self, project_id: str, filters: Optional[Dict[str, str]] = None) -> ItemList:
        """
        Returns a list of date settings models.
        """
        url = self.get_date_settings_models_url(project_id, filters)
        params = {'take': GET_LIMIT}
        date_settings = self._get_items(url, params)

        return self._keysort(date_settings, SORT_ORDER)


    def get_date_settings_model_by_id(self, project_id: str, model_id: str) -> Item:
        """
        Returns a specific date settings model from its id.
        """
        url = self.get_date_settings_model_by_id_url(project_id, model_id)
        date_settings = self._get_items(url)

        return date_settings[0]


    def get_date_settings_assignments_by_id(self, project_id: str, model_id: str) -> Union[ItemList, None]:
        """
        Returns a specific date settings model assignment from its id.
        """
        url = self.get_date_settings_assignments_by_id_url(project_id, model_id)
        assignments = self._get_items(url)

        if len(assignments) == 0:
            return None

        return assignments


    def get_depreciation(self, project_id: str, filters: Optional[Dict[str, str]] = None) -> ItemList:
        """
        Returns a list of depreciation models.
        """
        url = self.get_depreciation_models_url(project_id, filters)
        params = {'take': GET_LIMIT}
        depreciation = self._get_items(url, params)

        return self._keysort(depreciation, SORT_ORDER)


    def get_depreciation_model_by_id(self, project_id: str, model_id: str) -> Item:
        """
        Returns a specific depreciation model from its id.
        """
        url = self.get_depreciation_model_by_id_url(project_id, model_id)
        depreciation = self._get_items(url)

        return depreciation[0]


    def get_depreciation_assignments_by_id(self, project_id: str, model_id: str) -> Union[ItemList, None]:
        """
        Returns a specific depreciation model assignment from its id.
        """
        url = self.get_depreciation_assignments_by_id_url(project_id, model_id)
        assignments = self._get_items(url)

        if len(assignments) == 0:
            return None

        return assignments



import json
from pathlib import Path
from typing import List, Dict, NamedTuple, Union
from typing_extensions import Self


USER_HOME = Path.home()
PACKAGE_ROOT = Path(__file__).parent.resolve()

COMBOCURVE_JSON = USER_HOME / '.combocurve' / 'combocurve.json'
CC_API_CONFIG_JSON = USER_HOME / '.combocurve' / 'cc-api.config.json'

REFRENCE_WELLHEADER: Dict[str, Union[str, int, float, bool]] = json.loads(
    (PACKAGE_ROOT / 'assets' / 'wellHeader.json').read_text())

ECON_MODELS: List[Dict[str, str]] = json.loads(
    (PACKAGE_ROOT / 'assets' / 'econModels.json').read_text())


class Configuration(NamedTuple):
    apikey: str

    @classmethod
    def from_file(cls, path: Union[str, Path]) -> Self:
        """
        Parse the contents of a `cc-api.config.json` file from a path.
        """
        if isinstance(path, str):
            path = Path(path)

        doc = json.loads(path.read_text())

        return cls(
            apikey=doc['apikey'],
        )

# default to the configuration file bundled with the package
cfg = Configuration.from_file(CC_API_CONFIG_JSON)



import warnings
from typing import List, Dict, Optional, Union, Any, Iterator, Mapping

from .base import APIBase, Item, ItemList


GET_LIMIT = 200
GET_LIMIT_OUTPUTS_ARIES = 1000


class Forecasts(APIBase):
    ######
    # URLs
    ######

    def get_forecasts_url(self, project_id: str, filters: Optional[Dict[str, str]] = None) -> str:
        """
        Returns the API url of forecasts for a specific project id.
        """
        url = f'{self.API_BASE_URL}/projects/{project_id}/forecasts'
        if filters is None:
            return url

        url += self._build_params_string(filters)
        return url


    def get_forecast_by_id_url(self, project_id: str, forecast_id: str) -> str:
        """
        Returns the API url for a specific forecast from its forecast id.
        """
        base_url = self.get_forecasts_url(project_id)
        return f'{base_url}/{forecast_id}'


    def get_forecast_wells_url(self, project_id: str, forecast_id: str) -> str:
        """
        Returns the API url for a specific forecast's wells from its forecast id.
        """
        base_url = self.get_forecast_by_id_url(project_id, forecast_id)
        return f'{base_url}/wells'


    def get_forecast_aries_url(
            self, project_id: str, forecast_id: str, filters: Optional[Dict[str, str]] = None) -> str:
        """
        Returns the API url for a specific forecast's ARIES parameters from its forecast id.
        """
        base_url = self.get_forecast_by_id_url(project_id, forecast_id)
        url = f'{base_url}/aries'
        if filters is None:
            return url

        url += self._build_params_string(filters)
        return url


    def get_forecast_outputs_url(
            self, project_id: str, forecast_id: str, filters: Optional[Dict[str, str]] = None) -> str:
        """
        Returns the API url for a specific forecast outputs from its forecast id.
        """
        base_url = self.get_forecast_by_id_url(project_id, forecast_id)
        url = f'{base_url}/outputs'
        if filters is None:
            return url

        url += self._build_params_string(filters)
        return url


    def get_forecast_output_by_id_url(self, project_id: str, forecast_id: str, output_id: str) -> str:
        """
        Returns the API url for a specific forecast output from its forecast id and output id.
        """
        base_url = self.get_forecast_outputs_url(project_id, forecast_id)
        return f'{base_url}/{output_id}'


    def get_forecast_daily_volumes_url(
            self, project_id: str, forecast_id: str, filters: Optional[Dict[str, str]] = None) -> str:
        """
        Returns the API url for daily volumes for a specific project id and forecast id.
        """
        base_url = self.get_forecast_by_id_url(project_id, forecast_id)
        url = f'{base_url}/daily-volumes'
        if filters is None:
            return url

        url += self._build_params_string(filters)
        return url


    def get_forecast_monthly_volumes_url(
            self, project_id: str, forecast_id: str, filters: Optional[Dict[str, str]] = None) -> str:
        """
        Returns the API url for monthly volumes for a specific project id and forecast id.
        """
        base_url = self.get_forecast_by_id_url(project_id, forecast_id)
        url = f'{base_url}/monthly-volumes'
        if filters is None:
            return url

        url += self._build_params_string(filters)
        return url


    def get_forecast_segment_parameters_url(
            self, project_id: str, forecast_id: str, well_id: str, phase: str, series: str) -> str:
        """
        Returns the API url for a specific forecast segment parameters from its
        forecast id, well id, phase, and series.
        """

        VALID_PHASES = ['oil', 'gas', 'water']
        VALID_SERIES = ['best', 'p10', 'p50', 'p90']

        if phase.lower() not in VALID_PHASES:
            warnings.warn(f'`phase` \'{phase}\' is not in list of valid names:\n{VALID_PHASES}')

        if series.lower() not in VALID_SERIES:
            warnings.warn(f'`series` \'{series}\' is not in list of valid names:\n{VALID_SERIES}')

        phase = phase.lower()
        series = series.lower()

        base_url = self.get_forecast_by_id_url(project_id, forecast_id)
        url = f'{base_url}/parameters/{well_id}/{phase}/{series}'

        return url


    ###########
    # API calls
    ###########


    def get_forecasts(self, project_id: str, filters: Optional[Dict[str, str]] = None) -> ItemList:
        """
        Returns a list of forecasts for a specific project id.

        https://docs.api.combocurve.com/#3f80e303-3285-4ce1-aec8-40ab2d518e9e

        Example response:
        [
            {
                "createdAt": "2020-01-21T16:58:08.986Z",
                "id": "5e5981b9e23dae0012624d72",
                "name": "Test forecast",
                "runDate": "2020-06-08T19:14:22.012Z",
                "running": false,
                "type": "probabilistic",
                "updatedAt": "2020-01-21T17:58:08.986Z"
            }
        ]
        """
        url = self.get_forecasts_url(project_id, filters)
        params = {'take': GET_LIMIT}
        forecasts = self._get_items(url, params)

        order = {
            'name': 0,
            'id': 3,
            'createdAt': 2,
            'updatedAt': 1,
        }
        return self._keysort(forecasts, order)


    def post_forecasts(self, project_id: str, data: ItemList) -> ItemList:
        """
        Creates new forecasts for a specific project id.

        https://docs.api.combocurve.com/#e49824bc-678f-4b10-aaee-14d5ec1825cf

        Example data:
        [
            {
                "name": "Test forecast",
                "type": "deterministic"
            },
            {
                "name": "Test forecast 2",
                "type": "probabilistic"
            }
        ]

        Example response:
        {
            "generalErrors": [
                {
                    "name": "ValidationError",
                    "message": "The field 'name' is required.",
                    "location": "In body of request at position [0]"
                },
                {
                    "name": "ValidationError",
                    "message": "The field 'unique' is required.",
                    "location": "In body of request at position [2]"
                }
            ],
            "results": [
                {
                    "status": "Success",
                    "code": 200,
                    "name": "Test forecast"
                },
                {
                    "status": "Success",
                    "code": 200,
                    "name": "Test forecast 2"
                }
            ],
            "failedCount": 2,
            "successCount": 2
        }
        """
        url = self.get_forecasts_url(project_id)
        forecasts = self._post_items(url, data)

        return forecasts


    def get_forecast_by_id(self, project_id: str, forecast_id: str) -> Item:
        """
        Returns a specific forecast from its forecast id.

        https://docs.api.combocurve.com/#5fec685c-a64e-43bb-b32a-920d1a0c3be7

        Example response:
        {
            "createdAt": "2020-01-21T16:58:08.986Z",
            "id": "5e5981b9e23dae0012624d72",
            "name": "Test forecast",
            "runDate": "2020-06-08T19:14:22.012Z",
            "running": false,
            "type": "probabilistic",
            "updatedAt": "2020-01-21T17:58:08.986Z"
        }
        """
        url = self.get_forecast_by_id_url(project_id, forecast_id)
        params = {'take': GET_LIMIT}
        forecasts = self._get_items(url, params)

        return forecasts[0]


    def get_forecast_aries(
            self, project_id: str, forecast_id: str, filters: Optional[Dict[str, str]] = None) -> ItemList:
        """
        Returns a list of ARIES parameters for a specific project id and forecast id.

        https://docs.api.combocurve.com/#d24c7390-9c9e-48e7-91aa-1ad2c9551bd2

        Example response:
        [
            {
                "well": "602d31de4477b92029913e56",
                "forecast": [
                    {
                        "PROPNUM": "1PDPOIL020",
                        "WELL NAME": "1PDPOIL235",
                        "WELL NUMBER": "1H",
                        "INPT ID": "DTHrr7FMza",
                        "API10": "1234567890",
                        "API12": "12345678901",
                        "API14": "12345678901214",
                        "CHOSEN ID": "1PDPOIL235",
                        "ARIES ID": "1PDPOIL235",
                        "PHDWIN ID": "9E5F5CC579867509254700023",
                        "SECTION": 4,
                        "SEQUENCE": 10,
                        "QUALIFIER": "CC_QUAL",
                        "KEYWORD": "CUMS",
                        "EXPRESSION": "0 0 0 0 0 0"
                    },
                    {
                        "PROPNUM": "1PDPOIL020",
                        "WELL NAME": "1PDPOIL236",
                        "WELL NUMBER": "2H",
                        "INPT ID": "KMLrr7FMqq",
                        "API10": "1234567890",
                        "API12": "12345678901",
                        "API14": "12345678901214",
                        "CHOSEN ID": "1PDPOIL236",
                        "ARIES ID": "1PDPOIL236",
                        "PHDWIN ID": "9E5F5CC579867509254700023",
                        "SECTION": 5,
                        "SEQUENCE": 20,
                        "QUALIFIER": "CC_QUAL",
                        "KEYWORD": "CUMS",
                        "EXPRESSION": "0 0 0 0 0 0"
                    },
                    {
                        "PROPNUM": "1PDPOIL020",
                        "WELL NAME": "1PDPOIL237",
                        "WELL NUMBER": "3H",
                        "INPT ID": "LGHrr9FMhg",
                        "API10": "1234567890",
                        "API12": "12345678901",
                        "API14": "12345678901214",
                        "CHOSEN ID": "1PDPOIL237",
                        "ARIES ID": "1PDPOIL237",
                        "PHDWIN ID": "9E5F5CC579867509254700023",
                        "SECTION": 6,
                        "SEQUENCE": 30,
                        "QUALIFIER": "CC_QUAL",
                        "KEYWORD": "CUMS",
                        "EXPRESSION": "0 0 0 0 0 0"
                    }
                ]
            }
        ]
        """
        url = self.get_forecast_aries_url(project_id, forecast_id, filters)
        params = {'take': GET_LIMIT_OUTPUTS_ARIES}
        return self._get_items(url, params)


    def get_forecast_outputs(
            self, project_id: str, forecast_id: str, filters: Optional[Dict[str, str]] = None) -> ItemList:
        """
        Returns a list of outputs for a specific project id and forecast id.

        https://docs.api.combocurve.com/#b396508a-bb08-4014-916e-3f7543dfd58b

        Example response:
        [
            {
                "best": {
                    "segments": [
                        {
                            "b": 1.5,
                            "diEffSec": 0.603721736666588,
                            "diNominal": 0.005491514853419089,
                            "endDate": "2001-02-12T00:00:00.000Z",
                            "qEnd": 0.10001738801851469,
                            "qStart": 2.2340862422997945,
                            "realizedDSwEffSec": 0.07999999999999996,
                            "segmentIndex": 1,
                            "segmentType": "arps_modified",
                            "startDate": "1982-05-15T00:00:00.000Z",
                            "swDate": "1989-07-17T02:09:12.931Z"
                        }
                    ],
                    "eur": 539853.3300559863
                },
                "createdAt": "2020-01-21T16:58:08.986Z",
                "forecasted": true,
                "forecastedAt": "2020-01-31T22:58:50.578Z",
                "forecastedBy": "5e272d39b78910dd2a1bd8ea",
                "id": "5e272f1d4b97ed0013313088",
                "forecast": "6285655e2448a0ec88951dba",
                "p10": {
                    "segments": [
                        {
                            "b": 1.5,
                            "diEffSec": 0.461422453817678,
                            "diNominal": 0.0027926834451796035,
                            "endDate": "2005-11-15T00:00:00.000Z",
                            "qEnd": 0.10000470857720545,
                            "qStart": 2.2340862422997945,
                            "realizedDSwEffSec": 0.07999999999999996,
                            "segmentIndex": 1,
                            "segmentType": "arps_modified",
                            "startDate": "1982-05-15T00:00:00.000Z",
                            "swDate": "1989-03-21T18:28:58.327Z"
                        }
                    ],
                    "eur": 539853.3300559863
                },
                "p50": {
                    "segments": [
                        {
                            "b": 1.5,
                            "diEffSec": 0.603721736666588,
                            "diNominal": 0.005491514853419089,
                            "endDate": "2001-02-12T00:00:00.000Z",
                            "qEnd": 0.10001738801851469,
                            "qStart": 2.2340862422997945,
                            "realizedDSwEffSec": 0.07999999999999996,
                            "segmentIndex": 1,
                            "segmentType": "arps_modified",
                            "startDate": "1982-05-15T00:00:00.000Z",
                            "swDate": "1989-07-17T02:09:12.931Z"
                        }
                    ]
                },
                "p90": {
                    "segments": [
                        {
                            "b": 1.4864126197388372,
                            "diEffSec": 0.6377483484117903,
                            "diNominal": 0.006490343943301474,
                            "endDate": "1999-10-06T00:00:00.000Z",
                            "qEnd": 0.1000051797269731,
                            "qStart": 2.2340862422997945,
                            "realizedDSwEffSec": 0.07999999999999996,
                            "segmentIndex": 1,
                            "segmentType": "arps_modified",
                            "startDate": "1982-05-15T00:00:00.000Z",
                            "swDate": "1989-08-30T11:51:00.367Z"
                        }
                    ],
                    "eur": 539853.3300559863
                },
                "phase": "oil",
                "reviewedAt": "2020-01-31T23:00:00.00Z",
                "reviewedBy": "5e272d38b78910dd2a1bd6b5",
                "runDate": "2020-01-31T22:58:50.578Z",
                "status": "in_progress",
                "typeCurveData": {
                    "name": "applied-type-curve",
                    "type": "ratio"
                },
                "updatedAt": "2020-01-21T17:58:08.986Z",
                "well": "5e272d39b78910dd2a1bd8fe",
                "data_freq": "monthly"
            },
            {
                "best": {
                    "segments": [
                        {
                            "b": 0.5871650672699218,
                            "diEffSec": 0.5717582558912915,
                            "diNominal": 0.003009175917809707,
                            "endDate": "2015-06-16T00:00:00.000Z",
                            "qEnd": 92.20980498278352,
                            "qStart": 423.3332616779596,
                            "realizedDSwEffSec": 0.07999999999999996,
                            "segmentIndex": 1,
                            "segmentType": "arps_modified",
                            "startDate": "2013-03-19T00:00:00.000Z",
                            "swDate": "2031-08-05T06:08:36.544Z"
                        }
                    ],
                    "eur": 539853.3300559863
                },
                "createdAt": "2020-01-21T16:58:08.986Z",
                "forecasted": true,
                "forecastedAt": "2020-01-31T22:58:50.578Z",
                "forecastedBy": "5e272d39b78910dd2a1bd8ea",
                "id": "5f24af6537794900129b17af",
                "forecast": "6285655e2448a0ec88951dba",
                "phase": "oil",
                "reviewedAt": "2020-01-31T23:00:00.00Z",
                "reviewedBy": "5e272d38b78910dd2a1bd6b5",
                "runDate": "2020-07-31T23:57:20.352Z",
                "status": "in_progress",
                "updatedAt": "2020-01-21T17:58:08.986Z",
                "well": "5ed67b2ec3629ac71ded298b",
                "data_freq": "monthly"
            },
            {
                "createdAt": "2020-01-21T16:58:08.986Z",
                "forecasted": true,
                "forecastedAt": "2020-01-31T22:58:50.578Z",
                "forecastedBy": "5e272d39b78910dd2a1bd8ea",
                "id": "5f24af6537794900129b17af",
                "forecast": "6285655e2448a0ec88951dba",
                "phase": "oil",
                "ratio": {
                    "basePhase": "oil",
                    "segments": [
                        {
                            "b": 0.5871650672699218,
                            "diEffSec": 0.5717582558912915,
                            "diNominal": 0.003009175917809707,
                            "endDate": "2015-06-16T00:00:00.000Z",
                            "qEnd": 92.20980498278352,
                            "qStart": 423.3332616779596,
                            "realizedDSwEffSec": 0.07999999999999996,
                            "segmentIndex": 1,
                            "segmentType": "arps_modified",
                            "startDate": "2013-03-19T00:00:00.000Z",
                            "swDate": "2031-08-05T06:08:36.544Z"
                        }
                    ],
                    "eur": 4544.45
                },
                "reviewedAt": "2020-01-31T23:00:00.00Z",
                "reviewedBy": "5e272d38b78910dd2a1bd6b5",
                "runDate": "2020-07-31T23:57:20.352Z",
                "status": "in_progress",
                "typeCurveData": {
                    "name": "applied-type-curve",
                    "type": "ratio"
                },
                "updatedAt": "2020-01-21T17:58:08.986Z",
                "well": "5ed67b2ec3629ac71ded298b",
                "data_freq": "monthly"
            }
        ]
        """
        url = self.get_forecast_outputs_url(project_id, forecast_id, filters)
        params = {'take': GET_LIMIT_OUTPUTS_ARIES}
        return self._get_items(url, params)


    def get_forecast_output_by_id(self, project_id: str, forecast_id: str, output_id: str) -> Item:
        """
        Returns a specific forecast output from its forecast id and output id.

        https://docs.api.combocurve.com/#4a6312b3-51cf-48d8-bef2-7752e840b2e0

        Example response:
        {
            "best": {
                "segments": [
                    {
                        "b": 1.5,
                        "diEffSec": 0.603721736666588,
                        "diNominal": 0.005491514853419089,
                        "endDate": "2001-02-12T00:00:00.000Z",
                        "qEnd": 0.10001738801851469,
                        "qStart": 2.2340862422997945,
                        "realizedDSwEffSec": 0.07999999999999996,
                        "segmentIndex": 1,
                        "segmentType": "arps_modified",
                        "startDate": "1982-05-15T00:00:00.000Z",
                        "swDate": "1989-07-17T02:09:12.931Z"
                    }
                ],
                "eur": 539853.3300559863
            },
            "createdAt": "2020-01-21T16:58:08.986Z",
            "forecasted": true,
            "forecastedAt": "2020-01-31T22:58:50.578Z",
            "forecastedBy": "5e272d39b78910dd2a1bd8ea",
            "id": "5e272f1d4b97ed0013313088",
            "forecast": "6285655e2448a0ec88951dba",
            "p10": {
                "segments": [
                    {
                        "b": 1.5,
                        "diEffSec": 0.461422453817678,
                        "diNominal": 0.0027926834451796035,
                        "endDate": "2005-11-15T00:00:00.000Z",
                        "qEnd": 0.10000470857720545,
                        "qStart": 2.2340862422997945,
                        "realizedDSwEffSec": 0.07999999999999996,
                        "segmentIndex": 1,
                        "segmentType": "arps_modified",
                        "startDate": "1982-05-15T00:00:00.000Z",
                        "swDate": "1989-03-21T18:28:58.327Z"
                    }
                ],
                "eur": 539853.3300559863
            },
            "p50": {
                "segments": [
                    {
                        "b": 1.5,
                        "diEffSec": 0.603721736666588,
                        "diNominal": 0.005491514853419089,
                        "endDate": "2001-02-12T00:00:00.000Z",
                        "qEnd": 0.10001738801851469,
                        "qStart": 2.2340862422997945,
                        "realizedDSwEffSec": 0.07999999999999996,
                        "segmentIndex": 1,
                        "segmentType": "arps_modified",
                        "startDate": "1982-05-15T00:00:00.000Z",
                        "swDate": "1989-07-17T02:09:12.931Z"
                    }
                ]
            },
            "p90": {
                "segments": [
                    {
                        "b": 1.4864126197388372,
                        "diEffSec": 0.6377483484117903,
                        "diNominal": 0.006490343943301474,
                        "endDate": "1999-10-06T00:00:00.000Z",
                        "qEnd": 0.1000051797269731,
                        "qStart": 2.2340862422997945,
                        "realizedDSwEffSec": 0.07999999999999996,
                        "segmentIndex": 1,
                        "segmentType": "arps_modified",
                        "startDate": "1982-05-15T00:00:00.000Z",
                        "swDate": "1989-08-30T11:51:00.367Z"
                    }
                ],
                "eur": 539853.3300559863
            },
            "phase": "OIL",
            "reviewedAt": "2020-01-31T23:00:00.00Z",
            "reviewedBy": "5e272d38b78910dd2a1bd6b5",
            "runDate": "2020-01-31T22:58:50.578Z",
            "status": "in_progress",
            "typeCurveData": {
                "name": "applied-type-curve",
                "type": "ratio"
            },
            "updatedAt": "2020-01-21T17:58:08.986Z",
            "well": "5e272d39b78910dd2a1bd8fe",
            "data_freq": "monthly"
        }
        """
        url = self.get_forecast_output_by_id_url(project_id, forecast_id, output_id)
        params = {'take': GET_LIMIT}
        outputs = self._get_items(url, params)

        return outputs[0]


    def get_forecast_daily_volumes(
            self, project_id: str, forecast_id: str, filters: Optional[Dict[str, str]] = None) -> ItemList:
        """
        Returns a list of daily volumes for a specific project id and forecast id.

        https://docs.api.combocurve.com/#5c8407df-a92f-464a-976c-60fc610caafd

        Example response:
        [
            {
                "project": "63bdcdf1dc401f0012613185",
                "forecast": "63bdce14dc401f00126131a7",
                "well": "63bdcdf56782656f8aaad644",
                "resolution": "daily",
                "phases": [
                    {
                        "phase": "gas",
                        "forecastOutputId": "63bdce17dc401f0012614f34",
                        "series": [
                            {
                                "series": "best",
                                "startDate": "2020-03-15T00:00:00.000Z",
                                "endDate": "2025-03-15T00:00:00.000Z",
                                "eur": 1980202.8287785284,
                                "volumes": [
                                    0,
                                    1,
                                    2
                                ]
                            }
                        ]
                    },
                    {
                        "phase": "oil",
                        "forecastOutputId": "63bdce17dc401f0012614f34",
                        "series": [
                            {
                                "series": "best",
                                "startDate": "2020-03-15T00:00:00.000Z",
                                "endDate": "2025-03-15T00:00:00.000Z",
                                "eur": 610156.72736362,
                                "volumes": [
                                    0,
                                    1,
                                    2
                                ]
                            }
                        ]
                    }
                ]
            },
            {
                "project": "63bdcdf1dc401f0012613185",
                "forecast": "63bdce14dc401f00126131a7",
                "well": "63bdcdf56782654efaaad6ca",
                "resolution": "daily",
                "phases": [
                    {
                        "phase": "gas",
                        "forecastOutputId": "63bdce17dc401f00126150f5",
                        "series": [
                            {
                                "series": "best",
                                "startDate": "2020-03-15T00:00:00.000Z",
                                "endDate": "2025-03-15T00:00:00.000Z",
                                "eur": 1980202.8287785284,
                                "volumes": [
                                    0,
                                    1,
                                    2
                                ]
                            }
                        ]
                    },
                    {
                        "phase": "oil",
                        "forecastOutputId": "63bdce17dc401f00126150f5",
                        "series": [
                            {
                                "series": "best",
                                "startDate": "2020-03-15T00:00:00.000Z",
                                "endDate": "2025-03-15T00:00:00.000Z",
                                "eur": 610156.72736362,
                                "volumes": [
                                    0,
                                    1,
                                    2
                                ]
                            }
                        ]
                    }
                ]
            }
        ]
        """
        url = self.get_forecast_daily_volumes_url(project_id, forecast_id, filters)
        params = {'take': GET_LIMIT}
        daily_volumes = self._get_items(url, params)

        order = {
            'well': 0,
        }
        return self._keysort(daily_volumes, order)


    def get_forecast_monthly_volumes(
            self, project_id: str, forecast_id: str, filters: Optional[Dict[str, str]] = None) -> ItemList:
        """
        Returns a list of monthly volumes for a specific project id and forecast id.

        https://docs.api.combocurve.com/#f4d538aa-1c10-41f2-bcb3-d9e3716e2ecf

        Example response:
        [
            {
                "project": "63bdcdf1dc401f0012613185",
                "forecast": "63bdce14dc401f00126131a7",
                "well": "63bdcdf56782656f8aaad644",
                "resolution": "daily",
                "phases": [
                    {
                        "phase": "gas",
                        "forecastOutputId": "63bdce17dc401f0012614f34",
                        "series": [
                            {
                                "series": "best",
                                "startDate": "2020-03-15T00:00:00.000Z",
                                "endDate": "2025-03-15T00:00:00.000Z",
                                "eur": 1980202.8287785284,
                                "volumes": [
                                    0,
                                    1,
                                    2
                                ]
                            }
                        ]
                    },
                    {
                        "phase": "oil",
                        "forecastOutputId": "63bdce17dc401f0012614f34",
                        "series": [
                            {
                                "series": "best",
                                "startDate": "2020-03-15T00:00:00.000Z",
                                "endDate": "2025-03-15T00:00:00.000Z",
                                "eur": 610156.72736362,
                                "volumes": [
                                    0,
                                    1,
                                    2
                                ]
                            }
                        ]
                    }
                ]
            },
            {
                "project": "63bdcdf1dc401f0012613185",
                "forecast": "63bdce14dc401f00126131a7",
                "well": "63bdcdf56782654efaaad6ca",
                "resolution": "daily",
                "phases": [
                    {
                        "phase": "gas",
                        "forecastOutputId": "63bdce17dc401f00126150f5",
                        "series": [
                            {
                                "series": "best",
                                "startDate": "2020-03-15T00:00:00.000Z",
                                "endDate": "2025-03-15T00:00:00.000Z",
                                "eur": 1980202.8287785284,
                                "volumes": [
                                    0,
                                    1,
                                    2
                                ]
                            }
                        ]
                    },
                    {
                        "phase": "oil",
                        "forecastOutputId": "63bdce17dc401f00126150f5",
                        "series": [
                            {
                                "series": "best",
                                "startDate": "2020-03-15T00:00:00.000Z",
                                "endDate": "2025-03-15T00:00:00.000Z",
                                "eur": 610156.72736362,
                                "volumes": [
                                    0,
                                    1,
                                    2
                                ]
                            }
                        ]
                    }
                ]
            }
        ]
        """
        url = self.get_forecast_monthly_volumes_url(project_id, forecast_id, filters)
        params = {'take': GET_LIMIT}
        monthly_volumes = self._get_items(url, params)

        order = {
            'well': 0,
        }
        return self._keysort(monthly_volumes, order)


    def post_forecast_segment_parameters(
            self, project_id: str, forecast_id: str, well_id: str, phase: str, series: str,
            data: ItemList) -> ItemList:
        """
        Inserts a specific well's forecast parameters from its forecast id,
        well id, phase, and series.

        https://docs.api.combocurve.com/#498a95b7-5c23-4aae-b92a-191503ac0c5c

        Example data:
        [
            {
                "segmentType": "arps",
                "startDate": "2022-07-28",
                "endDate": "2028-07-26",
                "qStart": 497.54078888022735,
                "qEnd": 169.85103,
                "diEffSec": 0.2708,
                "b": 1.3
            },
            {
                "segmentType": "arps_modified",
                "startDate": "2028-07-27",
                "endDate": "2037-07-14",
                "qStart": 497.54078888022735,
                "qEnd": 108.2303,
                "diEffSec": 0.2708,
                "b": 0.9,
                "targetDSwEffSec": 0.06
            }
        ]

        Example response:
        {
            "status": "success",
            "segmentCount": "8,",
            "id": "62b1c13e2750169012ee4515"
        }
        """
        url = self.get_forecast_segment_parameters_url(project_id, forecast_id, well_id, phase, series)
        segments = self._post_items(url, data)

        return segments


    def put_forecast_segment_parameters(
            self, project_id: str, forecast_id: str, well_id: str, phase: str, series: str,
            data: ItemList) -> ItemList:
        """
        Updates a specific well's forecast parameters from its forecast id,
        well id, phase, and series.

        https://docs.api.combocurve.com/#59772626-e3af-458e-8ec4-9e57287706f6

        Example data:
        [
            {
                "segmentType": "arps",
                "startDate": "2022-07-28",
                "endDate": "2028-07-26",
                "qStart": 497.54078888022735,
                "qEnd": 169.85103,
                "diEffSec": 0.2708,
                "b": 1.3
            },
            {
                "segmentType": "arps_modified",
                "startDate": "2028-07-27",
                "endDate": "2037-07-14",
                "qStart": 497.54078888022735,
                "qEnd": 108.2303,
                "diEffSec": 0.2708,
                "b": 0.9,
                "targetDSwEffSec": 0.06
            }
        ]

        Example response:
        {
            "status": "success",
            "segmentCount": "8,",
            "id": "62b1c13e2750169012ee4515"
        }
        """
        url = self.get_forecast_segment_parameters_url(project_id, forecast_id, well_id, phase, series)
        segments = self._put_items(url, data)

        return segments


    def delete_forecast_segment_parameters(
            self, project_id: str, forecast_id: str, well_id: str, phase: str, series: str) -> ItemList:
        """
        Deletes a specific well's forecast parameters from its forecast id,
        well id, phase, and series.

        https://docs.api.combocurve.com/#943e863d-2dbd-4f83-a21e-f0305c072c07
        """
        url = self.get_forecast_segment_parameters_url(project_id, forecast_id, well_id, phase, series)
        segments = self._delete_items(url, data=[])

        return segments



import requests  # type: ignore
import warnings

from combocurve_api_v1.pagination import get_next_page_url  # type: ignore

from typing import List, Dict, Optional, Union, Any, Iterator, Mapping, cast

from .base import APIBase, Item, ItemList


GET_LIMIT = 200
GET_LIMIT_MONTHLY_EXPORTS = 100
CONCURRENCY_MONTHLY_EXPORTS = 10


def flatten_outputs(result: Item) -> Optional[Item]:
    if 'output' not in result:
        return None

    output = result.pop('output')

    if not isinstance(output, dict):
        raise TypeError(f'Expected output to be a dict, got {type(output)}')

    out = {k: v for k, v in output.items()}
    return {**result, **out}


class EconRuns(APIBase):
    ######
    # URLs
    ######

    def get_econ_runs_url(self, project_id: str, scenario_id: str) -> str:
        """
        Returns the API url of econ runs for a specific project id and
        scenario id.
        """
        return f'{self.API_BASE_URL}/projects/{project_id}/scenarios/{scenario_id}/econ-runs'


    def get_econ_run_by_id_url(self, project_id: str, scenario_id: str, econ_run_id: str) -> str:
        """
        Returns the API url for a specific econ run from its econ run id.
        """
        base_url = self.get_econ_runs_url(project_id, scenario_id)
        return f'{base_url}/{econ_run_id}'


    def get_econ_run_onelines_url(self, project_id: str, scenario_id: str, econ_run_id: str) -> str:
        """
        Returns the API url for onelines for a specific project id, scenario id,
        and econ run id.
        """
        base_url = self.get_econ_run_by_id_url(project_id, scenario_id, econ_run_id)
        return f'{base_url}/one-liners'


    def get_econ_run_combo_names_url(self, project_id: str, scenario_id: str, econ_run_id: str) -> str:
        """
        Returns the API url for onelines for a specific project id, scenario id,
        and econ run id.
        """
        base_url = self.get_econ_run_onelines_url(project_id, scenario_id, econ_run_id)
        return f'{base_url}/combo-names'


    def get_econ_run_monthly_export_id_url(self, project_id: str, scenario_id: str, econ_run_id: str) -> str:
        """
        Returns the API url for monthly exports for a specific project id,
        scenario id, and econ run id.
        """
        base_url = self.get_econ_run_by_id_url(project_id, scenario_id, econ_run_id)
        return f'{base_url}/monthly-exports'


    def get_econ_run_monthly_export_url(self, project_id: str, scenario_id: str, econ_run_id: str,
                                        monthly_export_id: str) -> str:
        """
        Returns the API url for monthly exports for a specific project id,
        scenario id, econ run id,
        and monthly export id.
        """
        base_url = self.get_econ_run_monthly_export_id_url(project_id, scenario_id, econ_run_id)
        return f'{base_url}/{monthly_export_id}'


    ###########
    # API calls
    ###########


    def get_econ_runs(self, project_id: str, scenario_id: str, add_combo_names: bool = True) -> ItemList:
        """
        Returns a list of econ runs for a specific project id and scenario id.

        `add_combo_names` will add the list of combo names to each econ run.
        """
        url = self.get_econ_runs_url(project_id, scenario_id)
        params = {'take': GET_LIMIT}
        econruns = self._get_items(url, params)

        if add_combo_names:
            self.update_econ_run_combo_names(econruns, project_id, scenario_id)

        order = {
            'id': 2,
            'status': 1,
            'runDate': 0,
        }
        return self._keysort(econruns, order, reverse=True)


    def get_econ_run_by_id(self, project_id: str, scenario_id: str, econ_run_id: str,
                           add_combo_names: bool = True) -> Item:
        """
        Returns a specific econ run from its econ run id.

        `add_combo_names` will add the list of combo names to the econ run.
        """
        url = self.get_econ_run_by_id_url(project_id, scenario_id, econ_run_id)
        params = {'take': GET_LIMIT}
        econruns = self._get_items(url, params)

        if add_combo_names:
            self.update_econ_run_combo_names(econruns, project_id, scenario_id)

        order = {
            'id': 2,
            'status': 1,
            'runDate': 0,
        }
        return self._keysort(econruns, order, reverse=True)[0]


    def get_econ_run_combo_names(self, project_id: str, scenario_id: str, econ_run_id: str) -> List[str]:
        """
        Returns a list of combo names for a specific project id, scenario id,
        and econ run id.
        """
        url = self.get_econ_run_combo_names_url(project_id, scenario_id, econ_run_id)
        params = {'take': GET_LIMIT}
        data = self._get_items(url, params)

        return cast(List[str], sorted(set(data)))


    def get_econ_run_onelines(self, project_id: str, scenario_id: str, econ_run_id: str) -> ItemList:
        """
        Returns a list of onelines for a specific project id, scenario id,
        and econ run id.
        """
        url = self.get_econ_run_onelines_url(project_id, scenario_id, econ_run_id)

        # in this specific case we do some post-processing to "flatten" the results
        # into the expected type of: ItemList
        params = {'take': GET_LIMIT}
        items = self._get_items(url, params)

        onelines = [
            item for item in
            (flatten_outputs(item) for item in items)
            if item is not None
        ]

        return onelines


    def update_econ_run_combo_names(self, econruns: ItemList, project_id: str, scenario_id: str) -> None:
        """
        Add combo names to the econ run data.
        """
        for i, run in enumerate(econruns):
            econ_run_id = str(run['id'])
            combo_names = self.get_econ_run_combo_names(project_id, scenario_id, econ_run_id)
            run['comboNames'] = combo_names
            # _ = run.pop('outputParams')
            econruns[i] = run

        return


    def post_econ_run_monthly_export(self, project_id: str, scenario_id: str, econ_run_id: str) -> str:
        """
        Create a monthly export for a specific project id, scenario id,
        econ run id, and returns a monthly export id to get the results.
        """
        url = self.get_econ_run_monthly_export_id_url(project_id, scenario_id, econ_run_id)

        items = self._post_items(url, data=[])
        id_ = str(items[0]['id'])

        return id_


    def get_econ_run_monthly_export(
            self, project_id: str, scenario_id: str, econ_run_id: str, monthly_export_id: str) -> ItemList:
        """
        Returns a list of monthly exports for a specific project id,
        scenario id, econ run id, and monthly export id.
        """
        url = self.get_econ_run_monthly_export_url(project_id, scenario_id, econ_run_id, monthly_export_id)

        params = {
            'take': GET_LIMIT_MONTHLY_EXPORTS,
            'concurrency': CONCURRENCY_MONTHLY_EXPORTS,
        }
        items = self._get_items(url, params)

        results = cast(ItemList, items[0]['results'])

        results_flat = [
            result for result in
            (flatten_outputs(result) for result in results)
            if result is not None
        ]
        return results_flat


    def get_stream_econ_run_monthly_export(
            self, project_id: str, scenario_id: str, econ_run_id: str, monthly_export_id: str) -> Iterator[ItemList]:
        """
        Similar to `get_econ_run_monthly_export` but instead streams the data
        yielding chunks of 100 items at a time, where each item is a list of
        monthly exports for a specific project id, scenario id, econ run id,
        and monthly export id.
        """
        url = self.get_econ_run_monthly_export_url(project_id, scenario_id, econ_run_id, monthly_export_id)

        params = {
            'take': GET_LIMIT_MONTHLY_EXPORTS,
            'concurrency': CONCURRENCY_MONTHLY_EXPORTS,
        }
        iter_items = self._get_items_iterator(url, params)

        for items in iter_items:
            for item in items:
                results = cast(ItemList, item['results'])

                results_flat = [
                    result for result in
                    (flatten_outputs(result) for result in results)
                    if result is not None
                ]
                yield results_flat
